from transformers import Seq2SeqTrainer
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers.trainer import *
from dataclasses import dataclass
from torch.utils.data.distributed import DistributedSampler
from transformers import is_torch_tpu_available
from torch.cuda.amp import autocast
from tqdm import tqdm

#import ipdb

import collections

import logging

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)



@dataclass
class modifiedLabelSmoother:
    """
    Adds label-smoothing on a pre-computed output from a Transformers model.

    Args:
        epsilon (:obj:`float`, `optional`, defaults to 0.1):
            The label smoothing factor.
        ignore_index (:obj:`int`, `optional`, defaults to -100):
            The index in the labels to ignore when computing the loss.
    """

    epsilon: float = 0.1
    fairseq: bool = False
    ignore_index: int = -100

    def __call__(self, model_output, labels):
        logits = model_output["logits"] if isinstance(model_output, dict) else model_output[0]
        log_probs = -nn.functional.log_softmax(logits, dim=-1)
        if labels.dim() == log_probs.dim() - 1:
            labels = labels.unsqueeze(-1)

        padding_mask = labels.eq(self.ignore_index)
        # In case the ignore_index is -100, the gather will fail, so we replace labels by 0. The padding_mask
        # will ignore them in any case.
        labels = torch.clamp(labels, min=0)
        labels = labels.to(torch.int64)             # I had to add this line to cast to int64, originally not in the code

        nll_loss = log_probs.gather(dim=-1, index=labels)
        # works for fp16 input tensor too, by internally upcasting it to fp32
        smoothed_loss = log_probs.sum(dim=-1, keepdim=True, dtype=torch.float32)

        nll_loss.masked_fill_(padding_mask, 0.0)
        smoothed_loss.masked_fill_(padding_mask, 0.0)

        num_active_elements = padding_mask.numel() - padding_mask.long().sum()

        ## fairseq implementation 
        if self.fairseq:
            nll_loss = nll_loss.sum()
            eps_i = self.epsilon / (log_probs.size(-1) - 1)
            loss = (1.0 - self.epsilon - eps_i) * nll_loss + eps_i * smoothed_loss.sum()
            return loss / num_active_elements, nll_loss / num_active_elements

        else:
        # huggingface implementation
        # Take the mean over the label dimensions, then divide by the number of active elements (i.e. not-padded):
            nll_loss = nll_loss.sum() / num_active_elements
            smoothed_loss = smoothed_loss.sum() / (num_active_elements * log_probs.shape[-1])
            return (1 - self.epsilon) * nll_loss + self.epsilon * smoothed_loss, nll_loss

# @dataclass
class modifiedUNlikelihood:

    def __init__(self, alpha = 0.1, prefix_allowed_tokens_fn = None):
        self.alpha = alpha
        self.prefix_allowed_tokens_fn = prefix_allowed_tokens_fn

    def __call__(self, model_output, labels):
        logits = model_output["logits"] if isinstance(model_output, dict) else model_output[0]
        unlikelihood = -torch.log(1-nn.functional.softmax(logits, dim=-1).clamp(min=0.0, max=0.9999))
        loss = 0
        count = 0

        for batch_i, sent in enumerate(labels):
            sent, anchor_idx = self.get_label(sent)
            sent = torch.tensor([16]+list(sent.cpu()))
            unlikhod_toks = self.prefix_allowed_tokens_fn(batch_i, torch.tensor([16]))
            i = 1
            while len(unlikhod_toks) > 1:
                unlikhod_toks.remove(sent[i])
                loss += torch.max(unlikelihood[batch_i, anchor_idx+i-1, unlikhod_toks])
                i += 1
                count += 1#len(unlikhod_toks)
                unlikhod_toks = self.prefix_allowed_tokens_fn(batch_i, sent[:i])
        loss = self.alpha * loss / count
        return loss

    def get_label(self, labels):
        valid = labels.ne(-100)
        return labels[valid], torch.min(torch.where(valid)[0])

def compute_kl_loss(p, q, pad_mask=None):
    log_p = p["logits"] if isinstance(p, dict) else p[0]
    log_q = q["logits"] if isinstance(q, dict) else q[0]

    p_loss = F.kl_div(F.log_softmax(log_p, dim=-1), F.softmax(log_q, dim=-1), reduction='none').sum(-1)
    q_loss = F.kl_div(F.log_softmax(log_q, dim=-1), F.softmax(log_p, dim=-1), reduction='none').sum(-1)

    # pad_mask is for seq-level tasks
    # import ipdb; ipdb.set_trace()
    if pad_mask is not None:
        p_loss.masked_fill_(pad_mask, 0.)
        q_loss.masked_fill_(pad_mask, 0.)

    if pad_mask is None:
        num_active = torch.ones_like(p_loss)
    else:
        num_active = torch.ones_like(pad_mask).sum() - pad_mask.float().sum()

    if num_active > 0.:
        p_loss = p_loss.sum() / num_active
        q_loss = q_loss.sum() / num_active

        loss = (p_loss + q_loss) / 2
    else:
        loss = 0.

    return loss

class modifiedSeq2SeqTrainer(Seq2SeqTrainer):

    def __init__(self, fairseq_loss = False, use_amp = False, enc_num = 0, dec_num = 0,prefix_allowed_tokens_fn=None,unlikelihood_weight=0.1, rdrop=0.0, generation_config =None, **kwargs):     # I had to add the use_amp arg and scaler

        super().__init__(**kwargs)
        self.label_smoother = modifiedLabelSmoother(epsilon=self.args.label_smoothing_factor, fairseq = fairseq_loss)
        if prefix_allowed_tokens_fn:
            self.unlikelihood_loss = modifiedUNlikelihood(alpha = unlikelihood_weight, prefix_allowed_tokens_fn=prefix_allowed_tokens_fn)
        else:
            self.unlikelihood_loss = None
        self.n_tokens_dec = dec_num
        self.n_tokens_enc = enc_num
        self.rdrop = rdrop
        self.use_amp = use_amp
    
    def reform_input(self, inputs, eos_token):

        enc_max_idx = torch.max(torch.where(inputs['input_ids']==eos_token)[1]) + 1
        label_max_idx = torch.max(torch.where(inputs['labels']==eos_token)[1]) + 1
        dec_max_idx = label_max_idx - self.n_tokens_dec
        inputs['input_ids'] = inputs['input_ids'][:,:enc_max_idx]
        inputs['decoder_input_ids'] = inputs['decoder_input_ids'][:,:dec_max_idx]
        inputs['attention_mask'] = inputs['attention_mask'][:,:enc_max_idx]
        inputs['decoder_attention_mask'] = inputs['decoder_attention_mask'][:,:dec_max_idx]
        inputs['labels'] = inputs['labels'][:,:label_max_idx]

        return inputs
        
    def compute_loss(self, model, inputs, return_outputs=False):
        inputs = self.reform_input(inputs, model.config.eos_token_id)
        if self.label_smoother is not None and "labels" in inputs:
            labels = inputs.pop("labels")
        else:
            labels = None
        outputs = model(**inputs)
        if self.rdrop > 0.:
            outputs2 = model(**inputs)
        # Save past state if it exists
        # TODO: this needs to be fixed and made cleaner later.
        if self.args.past_index >= 0:
            self._past = outputs[self.args.past_index]

        if labels is not None:
            loss, nll_loss = self.label_smoother(outputs, labels)
            if self.rdrop > 0.:
                loss2, nll_loss2 = self.label_smoother(outputs2, labels)
                loss = (loss + loss2) * 0.5
                nll_loss = (nll_loss + nll_loss2) * 0.5
        else:
            # We don't use .loss here since the model may return tuples instead of ModelOutput.
            loss = outputs["loss"] if isinstance(outputs, dict) else outputs[0]
            nll_loss = loss
        
        if self.unlikelihood_loss:
            # if not labels:
            #     labels = inputs.pop("labels")
            # loss += self.unlikelihood_loss(outputs, labels)
            if not self.rdrop > 0.:
                loss += self.unlikelihood_loss(outputs, labels)
            else:
                loss += 0.5 * (self.unlikelihood_loss(outputs, labels) + self.unlikelihood_loss(outputs2, labels))
        
        if self.rdrop > 0.:
            if labels is None:
                label_mask = None
            else:
                label_mask = labels.eq(-100)
            rdrop_loss = compute_kl_loss(outputs, outputs2, label_mask)
            loss += self.rdrop * rdrop_loss

        return (loss, outputs) if return_outputs else loss, nll_loss

    
    def training_step(self, model: nn.Module, inputs: Dict[str, Union[torch.Tensor, Any]]) -> torch.Tensor:

        model.train()
        inputs = self._prepare_inputs(inputs)

        if is_sagemaker_mp_enabled():
            scaler = self.scaler if self.use_amp else None
            loss_mb = smp_forward_backward(model, inputs, self.args.gradient_accumulation_steps, scaler=scaler)
            return loss_mb.reduce_mean().detach().to(self.args.device)

        if self.use_amp:
            with autocast():
                loss, nll_loss = self.compute_loss(model, inputs)
        else:
            loss, nll_loss = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training

        if self.args.gradient_accumulation_steps > 1 and not self.deepspeed:
            # deepspeed handles loss scaling by gradient_accumulation_steps in its `backward`
            loss = loss / self.args.gradient_accumulation_steps

        if self.use_amp:
            self.scaler.scale(loss).backward()
        elif self.use_apex:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        elif self.deepspeed:
            # loss gets scaled under gradient_accumulation_steps in deepspeed
            loss = self.deepspeed.backward(loss)
        else:
            loss.backward()

        # ipdb.set_trace()

        return loss.detach(), nll_loss.detach()


    def _maybe_log_save_evaluate(self, tr_loss, model, trial, epoch, ignore_keys_for_eval, nll_loss = 0):
        if self.control.should_log:
            logs: Dict[str, float] = {}

            # all_gather + mean() to get average loss over all processes
            tr_loss_scalar = self._nested_gather(tr_loss).mean().item()

            # reset tr_loss to zero
            tr_loss -= tr_loss

            logs["loss"] = round(tr_loss_scalar / (self.state.global_step - self._globalstep_last_logged), 4)
            logs["learning_rate"] = self._get_learning_rate()

            logs["nll_loss"] = round(float(nll_loss), 4)
            
            self._total_loss_scalar += tr_loss_scalar
            self._globalstep_last_logged = self.state.global_step
            self.store_flos()

            self.log(logs)

        metrics = None
        if self.control.should_evaluate:
            metrics = self.evaluate(ignore_keys=ignore_keys_for_eval)
            self._report_to_hp_search(trial, epoch, metrics)

        if self.control.should_save:
            self._save_checkpoint(model, trial, metrics=metrics)
            self.control = self.callback_handler.on_save(self.args, self.state, self.control)
        
    def train(
        self,
        resume_from_checkpoint: Optional[Union[str, bool]] = None,
        trial: Union["optuna.Trial", Dict[str, Any]] = None,
        ignore_keys_for_eval: Optional[List[str]] = None,
        **kwargs,
    ):

        resume_from_checkpoint = None if not resume_from_checkpoint else resume_from_checkpoint

        # memory metrics - must set up as early as possible
        self._memory_tracker.start()

        args = self.args

        self.is_in_train = True

        # do_train is not a reliable argument, as it might not be set and .train() still called, so
        # the following is a workaround:
        if args.fp16_full_eval and not args.do_train:
            self._move_model_to_device(self.model, args.device)

        if "model_path" in kwargs:
            resume_from_checkpoint = kwargs.pop("model_path")
            warnings.warn(
                "`model_path` is deprecated and will be removed in a future version. Use `resume_from_checkpoint` "
                "instead.",
                FutureWarning,
            )
        if len(kwargs) > 0:
            raise TypeError(f"train() received got unexpected keyword arguments: {', '.join(list(kwargs.keys()))}.")
        # This might change the seed so needs to run first.
        self._hp_search_setup(trial)

        # Model re-init
        model_reloaded = False
        if self.model_init is not None:
            # Seed must be set before instantiating the model when using model_init.
            set_seed(args.seed)
            self.model = self.call_model_init(trial)
            model_reloaded = True
            # Reinitializes optimizer and scheduler
            self.optimizer, self.lr_scheduler = None, None

        # Load potential model checkpoint
        if isinstance(resume_from_checkpoint, bool) and resume_from_checkpoint:
            resume_from_checkpoint = get_last_checkpoint(args.output_dir)
            if resume_from_checkpoint is None:
                raise ValueError(f"No valid checkpoint found in output directory ({args.output_dir})")

        if resume_from_checkpoint is not None:
            if not os.path.isfile(os.path.join(resume_from_checkpoint, WEIGHTS_NAME)):
                raise ValueError(f"Can't find a valid checkpoint at {resume_from_checkpoint}")

            logger.info(f"Loading model from {resume_from_checkpoint}).")

            if os.path.isfile(os.path.join(resume_from_checkpoint, CONFIG_NAME)):
                config = PretrainedConfig.from_json_file(os.path.join(resume_from_checkpoint, CONFIG_NAME))
                checkpoint_version = config.transformers_version
                if checkpoint_version is not None and checkpoint_version != __version__:
                    logger.warn(
                        f"You are resuming training from a checkpoint trained with {checkpoint_version} of "
                        f"Transformers but your current version is {__version__}. This is not recommended and could "
                        "yield to errors or unwanted behaviors."
                    )

            if args.deepspeed:
                # will be resumed in deepspeed_init
                pass
            else:
                # We load the model state dict on the CPU to avoid an OOM error.
                state_dict = torch.load(os.path.join(resume_from_checkpoint, WEIGHTS_NAME), map_location="cpu")
                # If the model is on the GPU, it still works!
                self._load_state_dict_in_model(state_dict)

                # release memory
                del state_dict

        # If model was re-initialized, put it on the right device and update self.model_wrapped
        if model_reloaded:
            if self.place_model_on_device:
                self._move_model_to_device(self.model, args.device)
            self.model_wrapped = self.model

        # Keeping track whether we can can len() on the dataset or not
        train_dataset_is_sized = isinstance(self.train_dataset, collections.abc.Sized)

        # Data loader and number of training steps
        train_dataloader = self.get_train_dataloader()

        # Setting up training control variables:
        # number of training epochs: num_train_epochs
        # number of training steps per epoch: num_update_steps_per_epoch
        # total number of training steps to execute: max_steps
        total_train_batch_size = args.train_batch_size * args.gradient_accumulation_steps * args.world_size
        if train_dataset_is_sized:
            num_update_steps_per_epoch = len(train_dataloader) // args.gradient_accumulation_steps
            num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
            if args.max_steps > 0:
                max_steps = args.max_steps
                num_train_epochs = args.max_steps // num_update_steps_per_epoch + int(
                    args.max_steps % num_update_steps_per_epoch > 0
                )
                # May be slightly incorrect if the last batch in the training datalaoder has a smaller size but it's
                # the best we can do.
                num_train_samples = args.max_steps * total_train_batch_size
            else:
                max_steps = math.ceil(args.num_train_epochs * num_update_steps_per_epoch)
                num_train_epochs = math.ceil(args.num_train_epochs)
                num_train_samples = len(self.train_dataset) * args.num_train_epochs
        else:
            # see __init__. max_steps is set when the dataset has no __len__
            max_steps = args.max_steps
            # Setting a very large number of epochs so we go as many times as necessary over the iterator.
            num_train_epochs = sys.maxsize
            num_update_steps_per_epoch = max_steps
            num_train_samples = args.max_steps * total_train_batch_size

        if DebugOption.UNDERFLOW_OVERFLOW in self.args.debug:
            if self.args.n_gpu > 1:
                # nn.DataParallel(model) replicates the model, creating new variables and module
                # references registered here no longer work on other gpus, breaking the module
                raise ValueError(
                    "Currently --debug underflow_overflow is not supported under DP. Please use DDP (torch.distributed.launch)."
                )
            else:
                debug_overflow = DebugUnderflowOverflow(self.model)  # noqa

        #delay_optimizer_creation = self.sharded_ddp is not None and self.sharded_ddp != ShardedDDPOption.SIMPLE
        delay_optimizer_creation = False
        if args.deepspeed:
            deepspeed_engine, optimizer, lr_scheduler = deepspeed_init(
                self, num_training_steps=max_steps, resume_from_checkpoint=resume_from_checkpoint
            )
            self.model = deepspeed_engine.module
            self.model_wrapped = deepspeed_engine
            self.deepspeed = deepspeed_engine
            self.optimizer = optimizer
            self.lr_scheduler = lr_scheduler
        elif not delay_optimizer_creation:
            self.create_optimizer_and_scheduler(num_training_steps=max_steps)

        self.state = TrainerState()
        self.state.is_hyper_param_search = trial is not None

        # Activate gradient checkpointing if needed
        if args.gradient_checkpointing:
            self.model.gradient_checkpointing_enable()

        model = self._wrap_model(self.model_wrapped)

        # for the rest of this function `model` is the outside model, whether it was wrapped or not
        if model is not self.model:
            self.model_wrapped = model

        if delay_optimizer_creation:
            self.create_optimizer_and_scheduler(num_training_steps=max_steps)

        # Check if saved optimizer or scheduler states exist
        self._load_optimizer_and_scheduler(resume_from_checkpoint)

        # important: at this point:
        # self.model         is the Transformers Model
        # self.model_wrapped is DDP(Transformers Model), Deepspeed(Transformers Model), etc.

        # Train!
        num_examples = (
            self.num_examples(train_dataloader) if train_dataset_is_sized else total_train_batch_size * args.max_steps
        )

        logger.info("***** Running training *****")
        logger.info(f"  Num examples = {num_examples}")
        logger.info(f"  Num Epochs = {num_train_epochs}")
        logger.info(f"  Instantaneous batch size per device = {args.per_device_train_batch_size}")
        logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_train_batch_size}")
        logger.info(f"  Gradient Accumulation steps = {args.gradient_accumulation_steps}")
        logger.info(f"  Total optimization steps = {max_steps}")

        self.state.epoch = 0
        start_time = time.time()
        epochs_trained = 0
        steps_trained_in_current_epoch = 0
        steps_trained_progress_bar = None

        # Check if continuing training from a checkpoint
        if resume_from_checkpoint is not None and os.path.isfile(
            os.path.join(resume_from_checkpoint, TRAINER_STATE_NAME)
        ):
            self.state = TrainerState.load_from_json(os.path.join(resume_from_checkpoint, TRAINER_STATE_NAME))
            epochs_trained = self.state.global_step // num_update_steps_per_epoch
            if not args.ignore_data_skip:
                steps_trained_in_current_epoch = self.state.global_step % (num_update_steps_per_epoch)
                steps_trained_in_current_epoch *= args.gradient_accumulation_steps
            else:
                steps_trained_in_current_epoch = 0

            logger.info("  Continuing training from checkpoint, will skip to saved global_step")
            logger.info(f"  Continuing training from epoch {epochs_trained}")
            logger.info(f"  Continuing training from global step {self.state.global_step}")
            if not args.ignore_data_skip:
                logger.info(
                    f"  Will skip the first {epochs_trained} epochs then the first {steps_trained_in_current_epoch} "
                    "batches in the first epoch. If this takes a lot of time, you can add the `--ignore_data_skip` "
                    "flag to your launch command, but you will resume the training on data already seen by your model."
                )
                if self.is_local_process_zero() and not args.disable_tqdm:
                    steps_trained_progress_bar = tqdm(total=steps_trained_in_current_epoch)
                    steps_trained_progress_bar.set_description("Skipping the first batches")

        # Update the references
        self.callback_handler.model = self.model
        self.callback_handler.optimizer = self.optimizer
        self.callback_handler.lr_scheduler = self.lr_scheduler
        self.callback_handler.train_dataloader = train_dataloader
        self.state.trial_name = self.hp_name(trial) if self.hp_name is not None else None
        if trial is not None:
            assignments = trial.assignments if self.hp_search_backend == HPSearchBackend.SIGOPT else trial
            self.state.trial_params = hp_params(assignments)
        else:
            self.state.trial_params = None
        # This should be the same if the state has been saved but in case the training arguments changed, it's safer
        # to set this after the load.
        self.state.max_steps = max_steps
        self.state.num_train_epochs = num_train_epochs
        self.state.is_local_process_zero = self.is_local_process_zero()
        self.state.is_world_process_zero = self.is_world_process_zero()

        # tr_loss is a tensor to avoid synchronization of TPUs through .item()
        tr_loss = torch.tensor(0.0).to(args.device)
        # _total_loss_scalar is updated everytime .item() has to be called on tr_loss and stores the sum of all losses
        self._total_loss_scalar = 0.0
        self._globalstep_last_logged = self.state.global_step
        model.zero_grad()

        self.control = self.callback_handler.on_train_begin(args, self.state, self.control)

        # Skip the first epochs_trained epochs to get the random state of the dataloader at the right point.
        if not args.ignore_data_skip:
            for epoch in range(epochs_trained):
                # We just need to begin an iteration to create the randomization of the sampler.
                for _ in train_dataloader:
                    break

        for epoch in range(epochs_trained, num_train_epochs):
            if isinstance(train_dataloader, DataLoader) and isinstance(train_dataloader.sampler, DistributedSampler):
                train_dataloader.sampler.set_epoch(epoch)
            elif isinstance(train_dataloader.dataset, IterableDatasetShard):
                train_dataloader.dataset.set_epoch(epoch)

            if is_torch_tpu_available():
                parallel_loader = pl.ParallelLoader(train_dataloader, [args.device]).per_device_loader(args.device)
                epoch_iterator = parallel_loader
            else:
                epoch_iterator = train_dataloader

            # Reset the past mems state at the beginning of each epoch if necessary.
            if args.past_index >= 0:
                self._past = None

            steps_in_epoch = (
                len(epoch_iterator) if train_dataset_is_sized else args.max_steps * args.gradient_accumulation_steps
            )
            self.control = self.callback_handler.on_epoch_begin(args, self.state, self.control)

            for step, inputs in enumerate(epoch_iterator):

                # Skip past any already trained steps if resuming training
                if steps_trained_in_current_epoch > 0:
                    steps_trained_in_current_epoch -= 1
                    if steps_trained_progress_bar is not None:
                        steps_trained_progress_bar.update(1)
                    if steps_trained_in_current_epoch == 0:
                        self._load_rng_state(resume_from_checkpoint)
                    continue
                elif steps_trained_progress_bar is not None:
                    steps_trained_progress_bar.close()
                    steps_trained_progress_bar = None

                if step % args.gradient_accumulation_steps == 0:
                    self.control = self.callback_handler.on_step_begin(args, self.state, self.control)

                if (
                    ((step + 1) % args.gradient_accumulation_steps != 0)
                    and args.local_rank != -1
                    and args._no_sync_in_gradient_accumulation
                ):
                    # Avoid unnecessary DDP synchronization since there will be no backward pass on this example.
                    with model.no_sync():
                        tr_loss_step, nll_loss = self.training_step(model, inputs)
                else:
                    tr_loss_step, nll_loss = self.training_step(model, inputs)

                if (
                    args.logging_nan_inf_filter
                    and not is_torch_tpu_available()
                    and (torch.isnan(tr_loss_step) or torch.isinf(tr_loss_step))
                ):
                    # if loss is nan or inf simply add the average of previous logged losses
                    tr_loss += tr_loss / (1 + self.state.global_step - self._globalstep_last_logged)
                else:
                    tr_loss += tr_loss_step

                self.current_flos += float(self.floating_point_ops(inputs))

                # Optimizer step for deepspeed must be called on every step regardless of the value of gradient_accumulation_steps
                if self.deepspeed:
                    self.deepspeed.step()

                if (step + 1) % args.gradient_accumulation_steps == 0 or (
                    # last step in epoch but step is always smaller than gradient_accumulation_steps
                    steps_in_epoch <= args.gradient_accumulation_steps
                    and (step + 1) == steps_in_epoch
                ):
                    # Gradient clipping
                    if args.max_grad_norm is not None and args.max_grad_norm > 0 and not self.deepspeed:
                        # deepspeed does its own clipping

                        if self.use_amp:
                            # AMP: gradients need unscaling
                            self.scaler.unscale_(self.optimizer)

                        if hasattr(self.optimizer, "clip_grad_norm"):
                            # Some optimizers (like the sharded optimizer) have a specific way to do gradient clipping
                            self.optimizer.clip_grad_norm(args.max_grad_norm)
                        elif hasattr(model, "clip_grad_norm_"):
                            # Some models (like FullyShardedDDP) have a specific way to do gradient clipping
                            model.clip_grad_norm_(args.max_grad_norm)
                        else:
                            # Revert to normal clipping otherwise, handling Apex or full precision
                            nn.utils.clip_grad_norm_(
                                amp.master_params(self.optimizer) if self.use_apex else model.parameters(),
                                args.max_grad_norm,
                            )

                    # Optimizer step
                    optimizer_was_run = True
                    if self.deepspeed:
                        pass  # called outside the loop
                    elif is_torch_tpu_available():
                        xm.optimizer_step(self.optimizer)
                    elif self.use_amp:
                        scale_before = self.scaler.get_scale()
                        self.scaler.step(self.optimizer)
                        self.scaler.update()
                        scale_after = self.scaler.get_scale()
                        optimizer_was_run = scale_before <= scale_after
                    else:
                        self.optimizer.step()

                    if optimizer_was_run and not self.deepspeed:
                        self.lr_scheduler.step()

                    self.model.zero_grad()
                    self.state.global_step += 1
                    self.state.epoch = epoch + (step + 1) / steps_in_epoch
                    self.control = self.callback_handler.on_step_end(args, self.state, self.control)

                    self._maybe_log_save_evaluate(tr_loss, self.model, trial, epoch, ignore_keys_for_eval, nll_loss=nll_loss)
                else:
                    self.control = self.callback_handler.on_substep_end(args, self.state, self.control)

                if self.control.should_epoch_stop or self.control.should_training_stop:
                    break

            self.control = self.callback_handler.on_epoch_end(args, self.state, self.control)
            self._maybe_log_save_evaluate(tr_loss, model, trial, epoch, ignore_keys_for_eval, nll_loss=nll_loss)

            if DebugOption.TPU_METRICS_DEBUG in self.args.debug:
                if is_torch_tpu_available():
                    # tpu-comment: Logging debug metrics for PyTorch/XLA (compile, execute times, ops, etc.)
                    xm.master_print(met.metrics_report())
                else:
                    logger.warning(
                        "You enabled PyTorch/XLA debug metrics but you don't have a TPU "
                        "configured. Check your training configuration if this is unexpected."
                    )
            if self.control.should_training_stop:
                break

        if args.past_index and hasattr(self, "_past"):
            # Clean the state at the end of training
            delattr(self, "_past")

        logger.info("\n\nTraining completed. Do not forget to share your model on huggingface.co/models =)\n\n")
        if args.load_best_model_at_end and self.state.best_model_checkpoint is not None:
            # Wait for everyone to get here so we are sur the model has been saved by process 0.
            if is_torch_tpu_available():
                xm.rendezvous("load_best_model_at_end")
            elif args.local_rank != -1:
                dist.barrier()

            logger.info(
                f"Loading best model from {self.state.best_model_checkpoint} (score: {self.state.best_metric})."
            )

            best_model_path = os.path.join(self.state.best_model_checkpoint, WEIGHTS_NAME)
            if os.path.exists(best_model_path):
                # We load the model state dict on the CPU to avoid an OOM error.
                state_dict = torch.load(best_model_path, map_location="cpu")
                # If the model is on the GPU, it still works!
                self._load_state_dict_in_model(state_dict)
            else:
                logger.warn(
                    f"Could not locate the best model at {best_model_path}, if you are running a distributed training "
                    "on multiple nodes, you should activate `--save_on_each_node`."
                )

            if self.deepspeed:
                self.deepspeed.load_checkpoint(
                    self.state.best_model_checkpoint, load_optimizer_states=False, load_lr_scheduler_states=False
                )

        # add remaining tr_loss
        self._total_loss_scalar += tr_loss.item()
        train_loss = self._total_loss_scalar / self.state.global_step

        metrics = speed_metrics("train", start_time, num_samples=num_train_samples, num_steps=self.state.max_steps)
        self.store_flos()
        metrics["total_flos"] = self.state.total_flos
        metrics["train_loss"] = train_loss

        self.is_in_train = False

        self._memory_tracker.stop_and_update_metrics(metrics)

        self.log(metrics)

        self.control = self.callback_handler.on_train_end(args, self.state, self.control)

        return TrainOutput(self.state.global_step, train_loss, metrics)




from typing import Dict, Callable, Iterable, List, Optional
# from transformers.generation_logits_process import LogitsProcessor
# import torch
# import math
try:
    import marisa_trie
except ModuleNotFoundError:
    pass

# class PrefixConstrainedLogitsProcessor(LogitsProcessor):

#     def __init__(self, prefix_allowed_tokens_fn: Callable[[int, torch.Tensor], List[int]], num_beams: int):
#         self._prefix_allowed_tokens_fn = prefix_allowed_tokens_fn
#         self._num_beams = num_beams

#     def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
#         mask = torch.full_like(scores, -math.inf)
#         for batch_id, beam_sent in enumerate(input_ids.view(-1, self._num_beams, input_ids.shape[-1])):
#             for beam_id, sent in enumerate(beam_sent):
#                 mask[batch_id * self._num_beams + beam_id, self._prefix_allowed_tokens_fn(batch_id, sent)] = 0
#         return scores + mask




class Trie(object):
    def __init__(self, sequences: List[List[int]] = []):
        self.trie_dict = {}
        self.len = 0
        if sequences:
            for sequence in sequences:
                Trie._add_to_trie(sequence, self.trie_dict)
                self.len += 1

        self.append_trie = None
        self.bos_token_id = None

    def append(self, trie, bos_token_id):
        self.append_trie = trie
        self.bos_token_id = bos_token_id

    def add(self, sequence: List[int]):
        Trie._add_to_trie(sequence, self.trie_dict)
        self.len += 1

    def get(self, prefix_sequence: List[int]):
        return Trie._get_from_trie(
            prefix_sequence, self.trie_dict, self.append_trie, self.bos_token_id
        )

    @staticmethod
    def load_from_dict(trie_dict):
        trie = Trie()
        trie.trie_dict = trie_dict
        trie.len = sum(1 for _ in trie)
        return trie

    @staticmethod
    def _add_to_trie(sequence: List[int], trie_dict: Dict):
        if sequence:
            if sequence[0] not in trie_dict:
                trie_dict[sequence[0]] = {}
            Trie._add_to_trie(sequence[1:], trie_dict[sequence[0]])

    @staticmethod
    def _get_from_trie(
        prefix_sequence: List[int],
        trie_dict: Dict,
        append_trie=None,
        bos_token_id: int = None,
    ):
        if len(prefix_sequence) == 0:
            output = list(trie_dict.keys())
            if append_trie and bos_token_id in output:
                output.remove(bos_token_id)
                output += list(append_trie.trie_dict.keys())
            return output
        elif prefix_sequence[0] in trie_dict:
            return Trie._get_from_trie(
                prefix_sequence[1:],
                trie_dict[prefix_sequence[0]],
                append_trie,
                bos_token_id,
            )
        else:
            if append_trie:
                return append_trie.get(prefix_sequence)
            else:
                return []

    def __iter__(self):
        def _traverse(prefix_sequence, trie_dict):
            if trie_dict:
                for next_token in trie_dict:
                    yield from _traverse(
                        prefix_sequence + [next_token], trie_dict[next_token]
                    )
            else:
                yield prefix_sequence

        return _traverse([], self.trie_dict)

    def __len__(self):
        return self.len

    def __getitem__(self, value):
        return self.get(value)


class DummyTrieMention(object):
    def __init__(self, return_values):
        self._return_values = return_values

    def get(self, indices=None):
        return self._return_values


class DummyTrieEntity(object):
    def __init__(self, return_values, codes):
        self._return_values = list(
            set(return_values).difference(
                set(
                    codes[e]
                    for e in (
                        "start_mention_token",
                        "end_mention_token",
                        "start_entity_token",
                    )
                )
            )
        )
        self._codes = codes

    def get(self, indices, depth=0):
        if len(indices) == 0 and depth == 0:
            return self._codes["end_mention_token"]
        elif len(indices) == 0 and depth == 1:
            return self._codes["start_entity_token"]
        elif len(indices) == 0:
            return self._return_values
        elif len(indices) == 1 and indices[0] == self._codes["end_entity_token"]:
            return self._codes["EOS"]
        else:
            return self.get(indices[1:], depth=depth + 1)



import torch
import torch.nn.functional as F

@torch.no_grad()
def sequence_score(model, sample, len_penalty=1):
    """Score a batch of translations."""

    model.eval()
    hypos = []
    decoder_out = model(
                input_ids = sample['input_ids'],
                decoder_input_ids = sample['decoder_input_ids'],
                attention_mask = sample['attention_mask'],
                decoder_attention_mask = sample['decoder_attention_mask'],
                return_dict=True,
                output_attentions=False,
                output_hidden_states=False,
            )

    for b in range(sample['input_ids'].shape[0]):

        logit = decoder_out.logits[b, sample['anchor'][b]:sample['anchor'][b]+len(sample['labels'][b]), :]
        probs = F.log_softmax(logit, dim=-1, dtype=torch.float32)
        target_probs = probs.gather(-1, sample['labels'][b].reshape(-1,1))

        score_i = torch.sum(target_probs) / (len(target_probs)) ** len_penalty

        hypos.append(
            [
                {
                    "tokens": sample['labels'][b],
                    "score": score_i,
                    "positional_scores": target_probs,
                }
            ]
        )
    return hypos


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import List, Optional

import torch
import torch.nn as nn

from torch import Tensor


class Search(nn.Module):
    def __init__(self, tgt_dict):
        super().__init__()
        if tgt_dict is not None:
            self.pad = tgt_dict.pad()
            self.unk = tgt_dict.unk()
            self.eos = tgt_dict.eos()
            self.vocab_size = len(tgt_dict)
            self.src_lengths = torch.tensor(-1)
            self.supports_constraints = False
            self.stop_on_max_len = False
        else:
            self.pad = 1
            self.unk = None
            self.eos = 2
            self.vocab_size = 0
            self.src_lengths = 0
            self.supports_constraints = False
            self.stop_on_max_len = False

    def step(
        self, step, lprobs, scores, prev_output_tokens=None, original_batch_idxs=None
    ):
        """Take a single search step.

        Args:
            step: the current search step, starting at 0
            lprobs: (bsz x input_beam_size x vocab_size)
                the model's log-probabilities over the vocabulary at the current step
            scores: (bsz x input_beam_size x step)
                the historical model scores of each hypothesis up to this point
            prev_output_tokens: (bsz x step)
                the previously generated oputput tokens
            original_batch_idxs: (bsz)
                the tensor with the batch indices, in the range [0, bsz)
                this is useful in case there has been applied a re-ordering
                and we need to know the orignal indices

        Return: A tuple of (scores, indices, beams) where:
            scores: (bsz x output_beam_size)
                the scores of the chosen elements; output_beam_size can be
                larger than input_beam_size, e.g., we may return
                2*input_beam_size to account for EOS
            indices: (bsz x output_beam_size)
                the indices of the chosen elements
            beams: (bsz x output_beam_size)
                the hypothesis ids of the chosen elements, in the range [0, input_beam_size)
        """
        raise NotImplementedError

    @torch.jit.export
    def set_src_lengths(self, src_lengths):
        self.src_lengths = src_lengths

    @torch.jit.export
    def init_constraints(self, batch_constraints: Optional[Tensor], beam_size: int):
        """Initialize constraint states for constrained decoding (if supported).

        Args:
            batch_constraints: (torch.Tensor, optional)
                the list of constraints, in packed form
            beam_size: (int)
                the beam size
        Returns:
            *encoder_out* rearranged according to *new_order*
        """
        pass

    def prune_sentences(self, batch_idxs: Tensor):
        """
        Removes constraint states for completed sentences (if supported).
        This is called from sequence_generator._generate() when sentences are
        deleted from the batch.

        Args:
            batch_idxs: Indices of *sentences* whose constraint state should be *kept*.
        """
        pass

    def update_constraints(self, active_hypos: Tensor):
        """
        Updates the constraint states by selecting the beam items that are retained.
        This is called at each time step of sequence_generator._generate() when
        the set of 2 * {beam_size} candidate hypotheses are reduced to the beam size.

        Args:
            active_hypos: (batch size, beam size)
              list of integers denoting, for each sentence, which beam candidate items
              should be kept.
        """
        pass

class Sampling(Search):
    sampling_topk: int
    sampling_topp: float

    def __init__(self, tgt_dict, sampling_topk=-1, sampling_topp=-1.0):
        super().__init__(tgt_dict)
        self.sampling_topk = sampling_topk
        self.sampling_topp = sampling_topp

    def _sample_topp(self, lprobs):
        """Sample among the smallest set of elements whose cumulative probability mass exceeds p.

        See `"The Curious Case of Neural Text Degeneration"
        (Holtzman et al., 2019) <https://arxiv.org/abs/1904.09751>`_.

        Args:
            lprobs: (bsz x input_beam_size x vocab_size)
                the model's log-probabilities over the vocabulary at the current step

        Return: A tuple of (trimed_probs, truncated_indices) where:
            trimed_probs: (bsz x input_beam_size x ?)
                the model's probabilities over the elements selected to sample from. The
                width of the third dimension is determined by top-P.
            truncated_indices: (bsz x input_beam_size x ?)
                the indices of the chosen elements.
        """
        probs = lprobs.exp_()

        # sort the last dimension (vocab dimension) in descending order
        sorted_probs, sorted_indices = probs.sort(descending=True)

        # compute a mask to indicate the words to be included in the top-P set.
        cumsum_probs = sorted_probs.cumsum(dim=2)
        mask = cumsum_probs.lt(self.sampling_topp)

        # note that mask was computed by 'lt'. One more word needs to be included
        # so that the cumulative probability mass can exceed p.
        cumsum_mask = mask.cumsum(dim=2)
        last_included = cumsum_mask[:, :, -1:]
        last_included.clamp_(0, mask.size()[2] - 1)
        mask = mask.scatter_(2, last_included, 1)

        # truncate unnecessary dims.
        max_dim = last_included.max()
        truncated_mask = mask[:, :, : max_dim + 1]
        truncated_probs = sorted_probs[:, :, : max_dim + 1]
        truncated_indices = sorted_indices[:, :, : max_dim + 1]

        # trim the words that are not in top-P by setting their probabilities
        # to 0, so that they would not be sampled later.
        trim_mask = ~truncated_mask
        trimed_probs = truncated_probs.masked_fill_(trim_mask, 0)
        return trimed_probs, truncated_indices

    @torch.jit.export
    def step(
        self,
        step: int,
        lprobs,
        scores,
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()

        if self.sampling_topp > 0:
            # only sample from the smallest set of words whose cumulative probability mass exceeds p
            probs, top_indices = self._sample_topp(lprobs)
        elif self.sampling_topk > 0:
            # only sample from top-k candidates
            lprobs, top_indices = lprobs.topk(self.sampling_topk)
            probs = lprobs.exp_()
        else:
            probs = lprobs.exp_()

            # dummy data to be consistent with true branch for type check
            top_indices = torch.empty(0).to(probs)
        # sample
        if step == 0:
            indices_buf = torch.multinomial(
                probs.view(bsz, -1),
                beam_size,
                replacement=True,
            ).view(bsz, beam_size)
        else:
            indices_buf = torch.multinomial(
                probs.view(bsz * beam_size, -1),
                1,
                replacement=True,
            ).view(bsz, beam_size)

        if step == 0:
            # expand to beam size
            probs = probs.expand(bsz, beam_size, -1)

        # gather scores
        scores_buf = torch.gather(probs, dim=2, index=indices_buf.unsqueeze(-1))
        scores_buf = scores_buf.log_().view(bsz, -1)

        # remap indices if using top-k or top-P sampling
        if self.sampling_topk > 0 or self.sampling_topp > 0:
            indices_buf = torch.gather(
                top_indices.expand(bsz, beam_size, -1),
                dim=2,
                index=indices_buf.unsqueeze(-1),
            ).squeeze(2)

        if step == 0:
            beams_buf = indices_buf.new_zeros(bsz, beam_size)
        else:
            beams_buf = torch.arange(0, beam_size).to(indices_buf).repeat(bsz, 1)
            # make scores cumulative
            scores_buf.add_(
                torch.gather(scores[:, :, step - 1], dim=1, index=beams_buf)
            )

        return scores_buf, indices_buf, beams_buf


class BeamSearch(Search):
    def __init__(self, tgt_dict):
        super().__init__(tgt_dict)
        self.constraint_states = None

    @torch.jit.export
    def step(
        self,
        step: int,
        lprobs,
        scores: Optional[Tensor],
        prev_output_tokens: Optional[Tensor] = None,
        original_batch_idxs: Optional[Tensor] = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)

        top_prediction = torch.topk(
            lprobs.view(bsz, -1),
            k=min(
                # Take the best 2 x beam_size predictions. We'll choose the first
                # beam_size of these which don't predict eos to continue with.
                beam_size * 2,
                lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
            ),
        )
        scores_buf = top_prediction[0]
        indices_buf = top_prediction[1]
        # Project back into relative indices and beams
        beams_buf = indices_buf // vocab_size
        indices_buf = indices_buf.fmod(vocab_size)

        # At this point, beams_buf and indices_buf are single-dim and contain relative indices
        return scores_buf, indices_buf, beams_buf


class PrefixConstrainedBeamSearch(Search):
    def __init__(self, tgt_dict, prefix_allowed_tokens_fn):
        super().__init__(tgt_dict)
        self.prefix_allowed_tokens_fn = prefix_allowed_tokens_fn
        self.stop_on_max_len = True

    @torch.jit.export
    def apply_mask(self, x, prev_output_tokens, original_batch_idxs, prefix_mention_is=False, prefix_length = None, decoding_prefix = None):
        beam_size = x.shape[0] // original_batch_idxs.shape[0]
        original_batch_idxs = (
            original_batch_idxs.unsqueeze(-1).repeat((1, beam_size)).flatten().tolist()
        )

        mask = torch.full_like(x, -math.inf)
        for sent_i, (sent, batch_i) in enumerate(
            zip(prev_output_tokens, original_batch_idxs)
        ):
            if prefix_mention_is:
                if not decoding_prefix:
                    sent = sent[prefix_length:]
                    mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0
                else:
                    mask[sent_i, :, :] = 0    
            else:
                mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0

            # mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0

        return mask

    @torch.jit.export
    def step(
        self,
        step: int,
        lprobs: Tensor,
        scores: Tensor,
        prev_output_tokens: Tensor,
        original_batch_idxs: Tensor,
        prefix_mention_is: bool,
        prefix_length = None, 
        decoding_prefix = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        lprobs += self.apply_mask(
            lprobs.view(bsz * beam_size, 1, vocab_size),
            prev_output_tokens,
            original_batch_idxs,
            prefix_mention_is,
            prefix_length,
            decoding_prefix,
        ).view(bsz, beam_size, vocab_size)

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)

        top_prediction = torch.topk(
            lprobs.view(bsz, -1),
            k=min(
                # Take the best beam_size predictions. We'll choose the first
                # beam_size of these which don't predict eos to continue with.
                beam_size,
                lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
            ),
        )
        scores_buf = top_prediction[0]
        indices_buf = top_prediction[1]
        beams_buf = indices_buf // vocab_size
        indices_buf = indices_buf.fmod(vocab_size)
        return scores_buf, indices_buf, beams_buf


class PrefixConstrainedBeamSearchWithSampling(Search):
    def __init__(self, tgt_dict, prefix_allowed_tokens_fn, logit_thresholding = None):
        super().__init__(tgt_dict)
        self.logit_thresholding = logit_thresholding
        self.prefix_allowed_tokens_fn = prefix_allowed_tokens_fn
        self.stop_on_max_len = True

    @torch.jit.export
    def apply_mask(self, x, prev_output_tokens, original_batch_idxs, prefix_mention_is=False, prefix_length = None, decoding_prefix = None):
        beam_size = x.shape[0] // original_batch_idxs.shape[0]
        original_batch_idxs = (
            original_batch_idxs.unsqueeze(-1).repeat((1, beam_size)).flatten().tolist()
        )

        mask = torch.full_like(x, -math.inf)
        for sent_i, (sent, batch_i) in enumerate(
            zip(prev_output_tokens, original_batch_idxs)
        ):
            if prefix_mention_is:
                if not decoding_prefix:
                    sent = sent[prefix_length:]
                    mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0
                else:
                    mask[sent_i, :, :] = 0  
                # if 16 in sent:
                #     sent = sent[list(sent).index(16):]
                #     mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0
                # else:
                #     mask[sent_i, :, :] = 0      
            else:
                mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0

            # mask[sent_i, :, self.prefix_allowed_tokens_fn(batch_i, sent)] = 0

        return mask

    @torch.jit.export
    def step(
        self,
        step: int,
        lprobs: Tensor,
        scores: Tensor,
        prev_output_tokens: Tensor,
        original_batch_idxs: Tensor,
        prefix_mention_is: bool,
        prefix_length = None, 
        decoding_prefix = None,
    ):
        bsz, beam_size, vocab_size = lprobs.size()

        lprobs += self.apply_mask(
            lprobs.view(bsz * beam_size, 1, vocab_size),
            prev_output_tokens,
            original_batch_idxs,
            prefix_mention_is,
            prefix_length,
            decoding_prefix,
        ).view(bsz, beam_size, vocab_size)

        if self.logit_thresholding is not None and torch.min(lprobs) == -math.inf:
            mask = torch.full_like(lprobs, -20)
            mask[lprobs>self.logit_thresholding] = 0
            _, token_with_max_score = torch.topk(lprobs, k = 1)
            mask = mask.scatter(-1, token_with_max_score, 0)
            lprobs += mask

        if step == 0:
            # at the first step all hypotheses are equally likely, so use
            # only the first beam
            lprobs = lprobs[:, ::beam_size, :].contiguous()
        else:
            # make probs contain cumulative scores for each hypothesis
            assert scores is not None
            lprobs = lprobs + scores[:, :, step - 1].unsqueeze(-1)


        top_prediction = torch.topk(
            lprobs.view(bsz, -1),
            k=min(
                # Take the best beam_size predictions. We'll choose the first
                # beam_size of these which don't predict eos to continue with.
                beam_size,
                lprobs.view(bsz, -1).size(1) - 1,  # -1 so we never select pad
            ),
        )
        scores_buf = top_prediction[0]
        indices_buf = top_prediction[1]
        beams_buf = indices_buf // vocab_size
        indices_buf = indices_buf.fmod(vocab_size)
        return scores_buf, indices_buf, beams_buf






# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import torch
import sys

import numpy as np
from tqdm import tqdm
import blink.biencoder.data_process as data
from blink.common.params import ENT_START_TAG, ENT_END_TAG



def prepare_crossencoder_mentions(
    tokenizer,
    samples,
    max_context_length=32,
    mention_key="mention",
    context_key="context",
    ent_start_token=ENT_START_TAG,
    ent_end_token=ENT_END_TAG,
):

    context_input_list = []  # samples X 128

    for sample in tqdm(samples):
        context_tokens = data.get_context_representation(
            sample,
            tokenizer,
            max_context_length,
            mention_key,
            context_key,
            ent_start_token,
            ent_end_token,
        )
        tokens_ids = context_tokens["ids"]
        context_input_list.append(tokens_ids)

    context_input_list = np.asarray(context_input_list)
    return context_input_list


def prepare_crossencoder_candidates(
    tokenizer, labels, nns, id2title, id2text, max_cand_length=128, topk=100
):

    START_TOKEN = tokenizer.cls_token
    END_TOKEN = tokenizer.sep_token

    candidate_input_list = []  # samples X topk=10 X 128
    label_input_list = []  # samples
    idx = 0
    for label, nn in zip(labels, nns):
        candidates = []

        label_id = -1
        for jdx, candidate_id in enumerate(nn[:topk]):

            if label == candidate_id:
                label_id = jdx

            rep = data.get_candidate_representation(
                id2text[candidate_id],
                tokenizer,
                max_cand_length,
                id2title[candidate_id],
            )
            tokens_ids = rep["ids"]

            assert len(tokens_ids) == max_cand_length
            candidates.append(tokens_ids)

        label_input_list.append(label_id)
        candidate_input_list.append(candidates)

        idx += 1
        sys.stdout.write("{}/{} \r".format(idx, len(labels)))
        sys.stdout.flush()

    label_input_list = np.asarray(label_input_list)
    candidate_input_list = np.asarray(candidate_input_list)

    return label_input_list, candidate_input_list


def filter_crossencoder_tensor_input(
    context_input_list, label_input_list, candidate_input_list
):
    # remove the - 1 : examples for which gold is not among the candidates
    context_input_list_filtered = [
        x
        for x, y, z in zip(context_input_list, candidate_input_list, label_input_list)
        if z != -1
    ]
    label_input_list_filtered = [
        z
        for x, y, z in zip(context_input_list, candidate_input_list, label_input_list)
        if z != -1
    ]
    candidate_input_list_filtered = [
        y
        for x, y, z in zip(context_input_list, candidate_input_list, label_input_list)
        if z != -1
    ]
    return (
        context_input_list_filtered,
        label_input_list_filtered,
        candidate_input_list_filtered,
    )


def prepare_crossencoder_data(
    tokenizer, samples, labels, nns, id2title, id2text, keep_all=False
):

    # encode mentions
    context_input_list = prepare_crossencoder_mentions(tokenizer, samples)

    # encode candidates (output of biencoder)
    label_input_list, candidate_input_list = prepare_crossencoder_candidates(
        tokenizer, labels, nns, id2title, id2text
    )

    if not keep_all:
        # remove examples where the gold entity is not among the candidates
        (
            context_input_list,
            label_input_list,
            candidate_input_list,
        ) = filter_crossencoder_tensor_input(
            context_input_list, label_input_list, candidate_input_list
        )
    else:
        label_input_list = [0] * len(label_input_list)

    context_input = torch.LongTensor(context_input_list)
    label_input = torch.LongTensor(label_input_list)
    candidate_input = torch.LongTensor(candidate_input_list)

    return (
        context_input,
        candidate_input,
        label_input,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import os
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from tqdm import tqdm
from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME


from transformers import AutoTokenizer, AutoModel
from bioel.models.arboel.model.common.ranker_base import BertEncoder, get_model_obj
from bioel.models.arboel.model.common.optimizer import get_bert_optimizer
from bioel.models.arboel.model.common.params import (
    ENT_START_TAG,
    ENT_END_TAG,
    ENT_TITLE_TAG,
)


def load_crossencoder(params):
    # Init model
    crossencoder = CrossEncoderRanker(params)
    return crossencoder


class CrossEncoderModule(torch.nn.Module):
    def __init__(self, params):
        super(CrossEncoderModule, self).__init__()
        model_path = params["model_name_or_path"]
        encoder_model = AutoModel.from_pretrained(model_path)

        self.encoder = BertEncoder(
            encoder_model,
            params["out_dim"],
            layer_pulled=params["pull_from_layer"],
            add_linear=params["add_linear"],
        )
        self.config = self.encoder.bert_model.config

    def forward(
        self,
        token_idx_ctxt,
        segment_idx_ctxt,
        mask_ctxt,
    ):

        embedding_ctxt = self.encoder(token_idx_ctxt, segment_idx_ctxt, mask_ctxt)
        return embedding_ctxt.squeeze(-1)


class CrossEncoderRanker(torch.nn.Module):
    def __init__(self, params, shared=None):
        super(CrossEncoderRanker, self).__init__()
        self.params = params

        self.NULL_IDX = 0
        self.START_TOKEN = "[CLS]"
        self.END_TOKEN = "[SEP]"

        # init model
        self.build_model()
        # if params["path_to_model"] is not None:
        #     self.load_model(params["path_to_model"])

    def load_model(self, fname, cpu=False):
        if cpu:
            state_dict = torch.load(fname, map_location=lambda storage, location: "cpu")
        else:
            state_dict = torch.load(fname)
        self.model.load_state_dict(state_dict)

    def build_model(self):
        self.model = CrossEncoderModule(self.params)

    def save_model(self, output_dir):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        model_to_save = get_model_obj(self.model)
        output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
        output_config_file = os.path.join(output_dir, CONFIG_NAME)
        torch.save(model_to_save.state_dict(), output_model_file)
        model_to_save.config.to_json_file(output_config_file)

    def get_optimizer(self, optim_states=None, saved_optim_type=None):
        return get_bert_optimizer(
            [self.model],
            self.params["type_optimization"],
            self.params["learning_rate"],
            fp16=self.params.get("fp16"),
        )

    def score_candidate(self, text_vecs, context_len):
        num_cand = text_vecs.size(1)
        text_vecs = text_vecs.view(-1, text_vecs.size(-1))
        token_idx_ctxt, segment_idx_ctxt, mask_ctxt = to_bert_input(
            text_vecs,
            self.NULL_IDX,
            context_len,
        )

        embedding_ctxt = self.model(
            token_idx_ctxt,
            segment_idx_ctxt,
            mask_ctxt,
        )

        return embedding_ctxt.view(-1, num_cand)

    def forward(self, input_idx, label_input, context_len):
        scores = self.score_candidate(input_idx, context_len)
        # print("scores :", scores)
        # print("label_input :", label_input)
        loss = F.cross_entropy(scores, label_input, reduction="mean")
        return loss, scores


def to_bert_input(token_idx, null_idx, segment_pos):
    """token_idx is a 2D tensor int.
    return token_idx, segment_idx and mask
    """
    segment_idx = token_idx * 0
    if segment_pos > 0:
        segment_idx[:, segment_pos:] = token_idx[:, segment_pos:] > 0

    mask = token_idx != null_idx
    # nullify elements in case self.NULL_IDX was not 0
    # token_idx = token_idx * mask.long()
    return token_idx, segment_idx, mask



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import collections
import os
import io
import copy
import torch
import logging
import random
import time
import numpy as np
import pickle
import json
import ujson
from tqdm import tqdm, trange
from bioel.models.arboel.biencoder.model.common.optimizer import get_bert_optimizer
from bioel.models.arboel.biencoder.model.BiEncoderLightningModule import (
    create_versioned_filename,
)
from pytorch_transformers.optimization import WarmupLinearSchedule
from bioel.models.arboel.crossencoder.model.crossencoder import CrossEncoderRanker
import bioel.models.arboel.biencoder.data.data_process as data_process
from IPython import embed

from bioel.logger import setup_logger
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    dataset_to_df,
    add_deabbreviations,
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
)

import lightning as L
from bioel.models.arboel.biencoder.model.common.params import BlinkParser

# logger = setup_logger()
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME

OPTIM_SCHED_FNAME = "optim_sched.pth"


def convert_defaultdict(d):
    if isinstance(d, collections.defaultdict):
        d = {k: convert_defaultdict(v) for k, v in d.items()}
    elif isinstance(d, dict):
        return {k: convert_defaultdict(v) for k, v in d.items()}
    elif isinstance(d, np.int64):
        return int(d)
    elif isinstance(d, np.float64):
        return float(d)
    return d


def merge_dicts(dict1, dict2, non_summing_keys=None):
    """Merge two dictionaries with support for nested structures and summing numeric values"""
    if non_summing_keys is None:
        non_summing_keys = set()

    merged_dict = copy.deepcopy(dict1)
    for key, value in dict2.items():
        if key in merged_dict:
            if key in non_summing_keys:
                # Replace value for keys that should not be summed
                merged_dict[key] = copy.deepcopy(value)
            elif isinstance(value, collections.defaultdict) and isinstance(
                merged_dict[key], collections.defaultdict
            ):
                for sub_key, sub_value in value.items():
                    merged_dict[key][sub_key] = (
                        merged_dict[key].get(sub_key, 0) + sub_value
                    )
            elif isinstance(value, dict) and isinstance(merged_dict[key], dict):
                merged_dict[key] = merge_dicts(
                    merged_dict[key], value, non_summing_keys
                )
            elif isinstance(value, (int, float)) and isinstance(
                merged_dict[key], (int, float)
            ):
                merged_dict[key] += value
            else:
                if isinstance(value, list) and isinstance(merged_dict[key], list):
                    merged_dict[key].extend(value)
                else:
                    merged_dict[key] = copy.deepcopy(value)
        else:
            merged_dict[key] = copy.deepcopy(value)
    return merged_dict


def accuracy(out, labels, return_bool_arr=False):
    outputs = np.argmax(out, axis=1)
    if return_bool_arr:
        return outputs == labels, outputs
    return np.sum(outputs == labels)


def recall(out, labels, k, return_bool_arr=False):
    top_k_indices = np.argsort(-out, axis=1)[:, :k]
    matches = np.any(top_k_indices == labels[:, None], axis=1)
    if return_bool_arr:
        return matches, top_k_indices
    return matches


def evaluate_single_batch(
    reranker,
    batch,
    logger,
    context_length,
    params,
    output_eval=None,
    recall_k=None,
    mention_data=None,
    compute_macro_avg=False,
    store_failure_success=False,
):
    """
    Evaluate the performance of the crossencoder on a single batch of data.
    Parameters:
    - reranker: CrossEncoderRanker object
    - batch: tuple of torch tensors
    - logger: logger object
    - context_length: int
        Maximum context length for the crossencoder
    - params : dict
        Contains most of the relevant keys for training (embed_batch_size, batch_size, n_gpu, etc...)
    - output_eval: list (used for testing, not validation)
        List to store the output of the evaluation
    - recall_k: int
        Number of candidates to consider for recall
    - mention_data: (Optional) torch tensor
        Contains the test mention data (used for testing)
    - compute_macro_avg: bool
        Whether to compute the macro average accuracy (=Accuracy over each mention type).
    - store_failure_success: bool
        Whether to store the failure and success cases for analysis
    """
    context_input, label_input, mention_idxs = batch
    results = {}
    eval_accuracy = 0
    nb_eval_examples = 0

    if mention_data is not None:
        processed_mention_data = mention_data["mention_data"]
        dictionary = mention_data["entity_dictionary"]
        stored_candidates = mention_data["stored_candidates"]
        if store_failure_success:
            failsucc = {"failure": [], "success": []}

    n_evaluated_per_type = collections.defaultdict(int)
    n_hits_per_type = collections.defaultdict(int)

    with torch.no_grad():
        eval_loss, logits = reranker(context_input, label_input, context_length)

    logits = logits.detach().cpu().numpy()
    label_ids = label_input.cpu().numpy()

    # Accuracy
    tmp_eval_hits, predicted = accuracy(
        out=logits, labels=label_ids, return_bool_arr=True
    )
    eval_accuracy = np.sum(tmp_eval_hits)

    nb_eval_examples = context_input.size(0)

    if compute_macro_avg:
        for i, m_idx in enumerate(mention_idxs):
            mention_type = processed_mention_data[m_idx]["type"]
            n_evaluated_per_type[mention_type] += 1
            is_hit = tmp_eval_hits[i]
            n_hits_per_type[mention_type] += is_hit

    if store_failure_success:

        # Needed for the output for evaluation
        data = load_bigbio_dataset(params["dataset"])
        if params["path_to_abbrev"]:
            data_with_abbrev = add_deabbreviations(
                dataset=data, path_to_abbrev=params["path_to_abbrev"]
            )
        exclude = CUIS_TO_EXCLUDE[params["dataset"]]
        remap = CUIS_TO_REMAP[params["dataset"]]
        df = dataset_to_df(
            data_with_abbrev, entity_remapping_dict=remap, cuis_to_exclude=exclude
        )

        recall_tmp_eval_hits, recall_predicted = recall(
            out=logits, labels=label_ids, k=recall_k, return_bool_arr=True
        )
        eval_recall = np.sum(recall_tmp_eval_hits)
        print(f"Recall@{recall_k}: {eval_recall/nb_eval_examples*100}%")
        for i, m_idx in enumerate(mention_idxs):
            m_idx = m_idx.item()
            men_query = processed_mention_data[m_idx]
            print("mention_id : ", men_query["mention_id"])
            best_idx = recall_predicted[i][0]
            dict_pred = dictionary[stored_candidates["candidates"][m_idx][best_idx]]
            report_obj = {
                "mention_id": men_query["mention_id"],
                "mention_name": men_query["mention_name"],
                "mention_gold_cui": "|".join(men_query["label_cuis"]),
                "mention_gold_cui_name": "|".join(
                    [
                        dictionary[i]["title"]
                        for i in men_query["label_idxs"][: men_query["n_labels"]]
                    ]
                ),
                "predicted_name": dict_pred["title"],
                "predicted_cui": dict_pred["cui"],
            }
            failsucc["success" if tmp_eval_hits[i] else "failure"].append(report_obj)

            cuis_top_candidates = []
            mention_dict = {}
            for j in range(recall_k):
                idx = recall_predicted[i][j]
                dict_preds = dictionary[stored_candidates["candidates"][m_idx][idx]]
                cuis_top_candidates.append([dict_preds["cui"]])

            if ((df["mention_id"] + ".abbr_resolved") == men_query["mention_id"]).any():
                filtered_df = df[
                    df["mention_id"] + ".abbr_resolved" == men_query["mention_id"]
                ].copy()
                filtered_df["mention_id"] += ".abbr_resolved"

            else:
                filtered_df = df[df["mention_id"] == men_query["mention_id"]]

            if not filtered_df.empty:
                mention_dict = filtered_df.iloc[0].to_dict()
            else:
                print(
                    "This mention name was not found in the mention dataset :",
                    men_query["mention_id"],
                    men_query["mention_name"],
                )

            mention_dict["candidates"] = cuis_top_candidates

            if params["equivalent_cuis"]:
                synsets = ujson.load(
                    open(os.path.join(params["data_path"], "cui_synsets.json"))
                )
                mention_dict["candidates"] = [
                    synsets[y[0]] for y in mention_dict["candidates"]
                ]

            output_eval.append(mention_dict)

    results["nb_samples_evaluated"] = nb_eval_examples
    results["correct_pred"] = int(eval_accuracy)

    print(f"Eval: Accuracy: {eval_accuracy/nb_eval_examples*100}%")

    if compute_macro_avg:
        results["n_hits_per_type"] = n_hits_per_type
        results["n_evaluated_per_type"] = n_evaluated_per_type

    if store_failure_success:
        results["failure"] = failsucc["failure"]
        results["success"] = failsucc["success"]

    if not store_failure_success:
        logger.info(json.dumps(results))

    return results


class LitCrossEncoder(L.LightningModule):
    """ """

    def __init__(self, params):
        super(LitCrossEncoder, self).__init__()
        self.save_hyperparameters(params)
        self.reranker = CrossEncoderRanker(params)
        self.model = self.reranker.model

    def training_step(self, batch, batch_idx):
        context_input, label_input, _ = batch
        train_loss, _ = self.reranker(
            context_input, label_input, self.hparams["max_context_length"]
        )

        print("training loss :", train_loss)
        self.log(
            "train loss :", train_loss, on_step=True, on_epoch=True, sync_dist=True
        )

        return train_loss

    def configure_optimizers(self):
        # Define optimizer
        optimizer = get_bert_optimizer(
            models=[self.model],
            type_optimization=self.hparams["type_optimization"],
            learning_rate=self.hparams["learning_rate"],
            fp16=self.hparams.get("fp16"),
        )

        # Define scheduler
        num_train_steps = (
            int(
                len(self.trainer.datamodule.train_data["mention_data"])
                / self.hparams["train_batch_size"]
                / self.trainer.accumulate_grad_batches
            )
            * self.trainer.max_epochs
        )
        num_warmup_steps = int(num_train_steps * self.hparams["warmup_proportion"])

        scheduler = WarmupLinearSchedule(
            optimizer=optimizer,
            warmup_steps=num_warmup_steps,
            t_total=num_train_steps,
        )
        logger.info(" Num optimization steps = %d" % num_train_steps)
        logger.info(" Num warmup steps = %d", num_warmup_steps)
        return [optimizer], [{"scheduler": scheduler, "interval": "step"}]

    def validation_step(self, batch, batch_idx):

        results = evaluate_single_batch(
            reranker=self.reranker,
            batch=batch,
            logger=logger,
            context_length=self.hparams["max_context_length"],
            params=self.hparams,
        )

        self.log(
            "Accuracy",
            results["correct_pred"] / results["nb_samples_evaluated"],
            on_epoch=True,
            sync_dist=True,
        )

    def on_test_epoch_start(self):

        self.n_evaluated_per_type = {}
        self.n_mentions_per_type = {}
        self.output_eval = []

        self.test_results = {
            # Accuracy computed when we only considered mentions with the correct cui in top candidates
            "normalized_accuracy": 0,
            # Average accuracy across all types (Only mentions with the correct cui in top candidates).
            "normalized_macro_avg_acc": 0,
            # Number of samples with the correct cui in top candidates in the test set
            "filtered_length": int(
                len(self.trainer.datamodule.test_data["mention_data"])
                - self.trainer.datamodule.nb_no_label_test
            ),
            # Number of samples with and without the correct cui in top candidates in the test set
            "unfiltered_length": len(self.trainer.datamodule.test_data["mention_data"]),
            "n_hits_per_type": collections.defaultdict(int),
            "n_evaluated_per_type": collections.defaultdict(int),
        }

        if self.trainer.limit_test_batches == 1.0:
            # Accuracy computed when we included mentions without the correct cui in top candidates
            self.test_results["unnormalized_accuracy"] = 0
            # Average accuracy across all types (Including mentions without the correct cui in top candidates)
            self.test_results["unnormalized_macro_avg_acc"] = 0
            self.test_results["n_mentions_per_type"] = collections.defaultdict(int)

    def test_step(self, batch, batch_idx):

        results = evaluate_single_batch(
            reranker=self.reranker,
            batch=batch,
            logger=logger,
            context_length=self.hparams["max_context_length"],
            recall_k=self.hparams["recall_k"],
            params=self.hparams,
            output_eval=self.output_eval,
            mention_data=self.trainer.datamodule.test_data,
            compute_macro_avg=True,
            store_failure_success=True,
        )
        self.test_results = merge_dicts(self.test_results, results)

    def on_test_epoch_end(self):
        self.test_results["normalized_accuracy"] = float(
            self.test_results["correct_pred"]
            / self.test_results["nb_samples_evaluated"]
            * 100
        )

        for mention_type in self.test_results["n_evaluated_per_type"]:
            self.test_results["normalized_macro_avg_acc"] += float(
                self.test_results["n_hits_per_type"][mention_type]
                / self.test_results["n_evaluated_per_type"][mention_type]
                if self.test_results["n_evaluated_per_type"][mention_type] > 0
                else 0
            )

        self.test_results["normalized_macro_avg_acc"] = (
            self.test_results["normalized_macro_avg_acc"]
            / len(self.test_results["n_evaluated_per_type"])
            * 100
        )

        self.log(
            "Normalized Accuracy : Only considered mentions with the correct cui in top candidates",
            self.test_results["normalized_accuracy"],
            sync_dist=True,
        )

        if self.trainer.limit_test_batches == 1.0:

            self.test_results["unnormalized_accuracy"] = float(
                self.test_results["normalized_accuracy"]
                * self.test_results["filtered_length"]
                / self.test_results["unfiltered_length"]
            )

            self.log(
                "Unnormalized Accuracy : Include mentions without the correct cui in top candidates",
                self.test_results["unnormalized_accuracy"],
                sync_dist=True,
            )

            for men in self.trainer.datamodule.test_data["mention_data"]:
                self.test_results["n_mentions_per_type"][men["type"]] += 1

        self.test_results = convert_defaultdict(self.test_results)

        # Gather results from all GPUs
        gathered_test_results = [None] * torch.distributed.get_world_size()
        torch.distributed.all_gather_object(gathered_test_results, self.test_results)
        gathered_output_eval = [None] * torch.distributed.get_world_size()
        torch.distributed.all_gather_object(gathered_output_eval, self.output_eval)

        # Only the main process should save the combined results
        if self.trainer.is_global_zero:
            non_summing_keys = {
                "n_mentions_per_type",
                "filtered_length",
                "unfiltered_length",
            }
            combined_test_results = gathered_test_results[0]

            for result in gathered_test_results[1:]:
                combined_test_results = merge_dicts(
                    combined_test_results, result, non_summing_keys
                )

            # Average the results across all GPUs for the specific metrics
            world_size = torch.distributed.get_world_size()
            combined_test_results["normalized_accuracy"] /= world_size
            combined_test_results["normalized_macro_avg_acc"] /= world_size

            if self.trainer.limit_test_batches == 1.0 and isinstance(
                self.trainer.limit_test_batches, float
            ):
                combined_test_results["unnormalized_accuracy"] /= world_size

                # For unnormalized_macro_avg_acc
                for mention_type in combined_test_results["n_mentions_per_type"]:
                    combined_test_results["unnormalized_macro_avg_acc"] += float(
                        combined_test_results["n_hits_per_type"][mention_type]
                        / combined_test_results["n_mentions_per_type"][mention_type]
                        if combined_test_results["n_mentions_per_type"][mention_type]
                        > 0
                        else 0
                    )

                combined_test_results["unnormalized_macro_avg_acc"] = (
                    combined_test_results["unnormalized_macro_avg_acc"]
                    / len(combined_test_results["n_mentions_per_type"])
                    * 100
                )

            all_output_eval = [
                item for sublist in gathered_output_eval for item in sublist
            ]

            eval_filename = "crossencoder_eval_results"
            crossencoder_eval_results = create_versioned_filename(
                self.hparams["output_path"], eval_filename
            )
            with open(crossencoder_eval_results, "w") as f:
                json.dump(combined_test_results, f, indent=2)
                print(
                    f"\ncrossencoder_eval_results overview saved at: {crossencoder_eval_results}"
                )

            eval_filename2 = "crossencoder_output_eval"
            crossencoder_output_eval = create_versioned_filename(
                self.hparams["output_path"], eval_filename2
            )
            with open(crossencoder_output_eval, "w") as f:
                json.dump(all_output_eval, f, indent=2)
                print(
                    f"\ncrossencoder_output_eval overview saved at: {crossencoder_output_eval}"
                )

    def on_train_start(self):
        self.start_time = time.time()

    def on_train_end(self):
        if self.start_time is not None:
            execution_time = (time.time() - self.start_time) / 60
            logging.info(f"The training took {execution_time:.2f} minutes")
        else:
            logging.warning("Start time not set. Unable to calculate execution time.")

    def on_test_start(self):
        self.start_time = time.time()

    def on_test_end(self):
        if self.start_time is not None:
            execution_time = (time.time() - self.start_time) / 60
            logging.info(f"The testing took {execution_time:.2f} minutes")
        else:
            logging.warning("Start time not set. Unable to calculate execution time.")



#!/usr/bin/env python
import numpy as np
import argparse
import torch
import torch.optim as optim
from torch.cuda.amp import autocast
from torch.cuda.amp import GradScaler
from pytorch_metric_learning import samplers
import logging
import time
import pdb
import os
import json
import random
from tqdm import tqdm

import wandb

from bioel.models.sapbert.data.lightingDataModule import SapbertDataModule
from bioel.models.sapbert.model.model_wrapper import Model_Wrapper
from bioel.models.sapbert.model.metric_learning import Sap_Metric_Learning

LOGGER = logging.getLogger(__name__)


def parse_args(config):
    """
    Parse the input arguments
    """
    print(config)
    parser = argparse.ArgumentParser(description="SAPBERT Training")
    # Required
    parser.add_argument(
        "--model_dir",
        default=config["model_dir"],
        help="Directory for pretrained model",
    )
    parser.add_argument(
        "--train_dir",
        type=str,
        default=config["train_dir"],
        help="training set directory",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default=config["output_dir"],
        help="Directory for output",
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        choices=[
            "medmentions_full",
            "medmentions_st21pv",
            "bc5cdr",
            "nlmchem",
            "nlm_gene",
            "gnormplus",
            "ncbi_disease",
        ],
        default=config["dataset_name"],
        help="data set to evaluate",
    )
    parser.add_argument(
        "--dict_cache_path", type=str, default=config["dict_cache_path"]
    )
    # Tokenizer settings
    parser.add_argument("--max_length", type=int, default=config.get("max_length", 25))

    # Train config
    parser.add_argument("--use_cuda", default=config["use_cuda"])
    parser.add_argument(
        "--learning_rate",
        help="learning rate",
        default=config.get("learning_rate", 0.0001),
        type=float,
    )
    parser.add_argument(
        "--weight_decay",
        help="weight decay",
        default=config.get("weight_decay", 0.01),
        type=float,
    )
    parser.add_argument(
        "--train_batch_size",
        help="train batch size",
        default=config.get("train_batch_size", 240),
        type=int,
    )
    parser.add_argument(
        "--eval_batch_size",
        help="eval batch size",
        default=config.get("eval_batch_size", 240),
        type=int,
    )
    parser.add_argument(
        "--split",
        default=config.get("split", "test"),
        type=str,
        help="Which datasplit to evaluate model performance on",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Run model with small data subset as a test of functionality",
    )
    parser.add_argument(
        "--dict_batch_size", type=int, default=config.get("dict_batch_size", 8192)
    )
    parser.add_argument("--topk", type=int, default=config.get("topk", 10))

    parser.add_argument(
        "--epoch", help="epoch to train", default=config.get("epoch", 3), type=int
    )
    parser.add_argument(
        "--save_checkpoint_all", default=config.get("save_checkpoint_all", False)
    )
    parser.add_argument(
        "--checkpoint_step", type=int, default=config.get("checkpoint_step", 1000000)
    )
    parser.add_argument(
        "--amp",
        default=config.get("amp", False),
        help="automatic mixed precision training",
    )
    parser.add_argument("--parallel", default=config.get("parallel", False))
    # parser.add_argument('--cased', action="store_true")
    parser.add_argument(
        "--pairwise",
        default=config.get("pairwise", False),
        help="if loading pairwise formatted datasets",
    )
    parser.add_argument(
        "--random_seed",
        help="epoch to train",
        default=config.get("random_seed", 1992),
        type=int,
    )
    parser.add_argument(
        "--loss",
        help="{ms_loss|cosine_loss|circle_loss|triplet_loss}}",
        default=config.get("loss", "ms_loss"),
    )
    parser.add_argument("--use_miner", default=config.get("use_miner", False))
    parser.add_argument(
        "--miner_margin", default=config.get("miner_margin", 0.2), type=float
    )
    parser.add_argument(
        "--type_of_triplets", default=config.get("type_of_triplets", "all"), type=str
    )
    parser.add_argument(
        "--agg_mode",
        default=config.get("agg_mode", "cls"),
        type=str,
        help="{cls|mean|mean_all_tok}",
    )
    parser.add_argument(
        "--num_workers", default=config.get("num_workers", 16), type=int
    )
    parser.add_argument("--mode", default=config.get("mode", "pretrain"), type=str)
    parser.add_argument(
        "--project",
        default=config.get("project", "SAPBERT"),
        type=str,
        help="Wandb Project",
    )
    parser.add_argument(
        "--resolve_abbreviations",
        default=config.get("resolve_abbreviations", False),
        help="consider abbreviations for the BigBio",
    )
    parser.add_argument(
        "--path_to_abbreviation",
        default=config["path_to_abbreviation"],
        type=str,
        help="Path to the abbreviation dictionary",
    )

    args = parser.parse_args()
    return vars(args)


def init_logging():
    LOGGER.setLevel(logging.INFO)
    fmt = logging.Formatter("%(asctime)s: [ %(message)s ]", "%m/%d/%Y %I:%M:%S %p")
    console = logging.StreamHandler()
    console.setFormatter(fmt)
    LOGGER.addHandler(console)


def init_seed(seed=None):
    if seed is None:
        seed = int(round(time.time() * 1000)) % 10000

    LOGGER.info("Using seed={}, pid={}".format(seed, os.getpid()))
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True


def train(config, data_loader, model, scaler=None, model_wrapper=None, step_global=0):
    LOGGER.info("train!")

    train_loss = 0
    train_steps = 0
    model.cuda()
    model.train()
    for i, data in tqdm(enumerate(data_loader), total=len(data_loader)):
        model.optimizer.zero_grad()

        batch_x1, batch_x2, batch_y = data
        batch_x_cuda1, batch_x_cuda2 = {}, {}
        for k, v in batch_x1.items():
            batch_x_cuda1[k] = v.cuda()
        for k, v in batch_x2.items():
            batch_x_cuda2[k] = v.cuda()

        batch_y_cuda = batch_y.cuda()

        if config["amp"]:
            with autocast():
                loss = model(batch_x_cuda1, batch_x_cuda2, batch_y_cuda)
        else:
            loss = model(batch_x_cuda1, batch_x_cuda2, batch_y_cuda)
        if config["amp"]:
            scaler.scale(loss).backward()
            scaler.step(model.optimizer)
            scaler.update()
        else:
            loss.backward()
            model.optimizer.step()

        train_loss += loss.item()
        wandb.log({"Loss": loss.item()})
        train_steps += 1
        step_global += 1
        # if (i+1) % 10 == 0:
        # LOGGER.info ("epoch: {} loss: {:.3f}".format(i+1,train_loss / (train_steps+1e-9)))
        # LOGGER.info ("epoch: {} loss: {:.3f}".format(i+1, loss.item()))

        # save model every K iterations
        if step_global % config["checkpoint_step"] == 0:
            checkpoint_dir = os.path.join(
                config["output_dir"], "checkpoint_iter_{}".format(str(step_global))
            )
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)
            model_wrapper.save_model(checkpoint_dir)
    train_loss /= train_steps + 1e-9
    return train_loss, step_global


def train_model(params, model):
    model_wrapper = model  # model wrapper
    init_logging()
    torch.manual_seed(params["random_seed"])

    # Initialize Wandb
    wandb.init(project=params["project"], config=params)

    if not os.path.exists(params["output_dir"]):
        os.makedirs(params["output_dir"])

    # Load BERT Tokenizer, dense encoder
    # model_wrapper = Model_Wrapper()
    encoder, tokenizer = model_wrapper.load_bert(
        path=params["model_dir"],
        max_length=params["max_length"],
        use_cuda=params["use_cuda"],
    )

    # Load the SAP Model
    model_true = Sap_Metric_Learning(
        encoder=encoder,
        learning_rate=params["learning_rate"],
        weight_decay=params["weight_decay"],
        use_cuda=params["use_cuda"],
        pairwise=params["pairwise"],
        loss=params["loss"],
        use_miner=params["use_miner"],
        miner_margin=params["miner_margin"],
        type_of_triplets=params["type_of_triplets"],
        agg_mode=params["agg_mode"],
    )

    # paramsure for Parallel Processing
    if params["parallel"]:
        model_true.encoder = torch.nn.DataParallel(model_true.encoder)
        LOGGER.info("Parallel Processing enabled")

    # Prepare the date module
    data_module = SapbertDataModule(params)
    data_module.prepare_data()
    data_module.setup()
    train_loader = data_module.train_dataloader()

    # Mixed Precision Training
    if params["amp"]:
        scaler = GradScaler()
    else:
        scaler = None

    start = time.time()
    step_global = 0
    for epoch in range(1, params["epoch"] + 1):
        LOGGER.info("Epoch {}/{}".format(epoch, params["epoch"]))

        # train the model
        train_loss, step_global = train(
            params,
            data_loader=train_loader,
            model=model_true,
            scaler=scaler,
            model_wrapper=model_wrapper,
            step_global=step_global,
        )

        LOGGER.info("loss/train_per_epoch={}/{}".format(train_loss, epoch))

        if params["save_checkpoint_all"]:
            checkpoint_dir = os.path.join(
                params["output_dir"], "checkpoint_epoch_{}".format(str(epoch))
            )
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)
            model_wrapper.save_model(checkpoint_dir)

        # Save the model Last Epoch
        if epoch == params["epoch"]:
            model_wrapper.save_model(os.path.join(params["output_dir"], wandb.run.name))

    end = time.time()
    training_time = end - start
    training_hour = int(training_time / 60 / 60)
    training_min = int(training_time / 60 % 60)
    training_sec = int(training_time % 60)

    LOGGER.info(
        "Training Time: {} hours {} minutes {} seconds".format(
            training_hour, training_min, training_sec
        )
    )

    with open(os.path.join(params["output_dir"], "training_time.txt"), "w") as f:
        f.write(
            "Training Time: {} hours {} minutes {} seconds \n".format(
                training_hour, training_min, training_sec
            )
        )


# if __name__ == "__main__":
#     config = parse_args()
#     main(config)



import os
import pandas as pd
import logging


from tqdm import tqdm
from typing import Union, Optional
from logging import getLogger

tqdm.pandas()

logger = logging.getLogger(__file__)
formatter = logging.Formatter(
    "%(asctime)s [%(levelname)s] %(message)s", "%Y-%m-%d %H:%M:%S"
)

# from config import umls_api_key


class UmlsMappings(object):
    def __init__(
        self,
        umls_dir="/Users/david/umls/",
        umls_api_key="",
        force_reprocess=False,
        debug=False,
    ):
        self.umls_dir = umls_dir
        self.debug = debug
        self.force_reprocess = force_reprocess
        if not os.path.isfile(umls_dir + "MRCONSO.RRF"):
            self._download_umls(umls_api_key)

        self._load_umls()
        if self.using_cached:
            self._load_type_abbrevs()

        else:
            self._load_types()
            self._merge_types()
            self._add_definitions()

        # self.cui_to_aliases = self.get_aliases()
        # self.cui_to_types = self._process_types()
        # self.cui_to_groups = self._process_groups()

        self._cache_umls()

    def _download_umls(self, api_key):
        if api_key == "":
            raise ValueError(
                """Please add valid UTS API key to config.py! \
                (For instructions on how to obtain a UTS API key,\
                please see https://documentation.uts.nlm.nih.gov/automating-downloads.html"""
            )
        else:
            pass

    def _download_semtypes(self, api_key):
        """
        Download semantic type files from UMLS
        """
        # Semantic network files
        # https://lhncbc.nlm.nih.gov/semanticnetwork/download/sn_current.tgz

        # Semantic Group Files
        # https://lhncbc.nlm.nih.gov/semanticnetwork/download/SemGroups.txt
        pass

    def _add_definitions(self):
        '''
        Add definitions of entities where available
        '''
        umls_definitions = pd.read_csv(
            os.path.join(self.umls_dir, "MRDEF.RRF"),
            sep="|",
            usecols=[0,1,4,5,],
            names=["cui", "aui", "sab", 'def',],
        )

        self.umls = self.umls.join(umls_definitions.set_index(['cui','aui','sab']), on=['cui','aui','sab'])

    def _load_umls(self):
        """
        Load UMLS MRCONSO.RRF
        """
        self.using_cached = False

        cache_path = os.path.join(self.umls_dir, ".cached_df.feather")
        if os.path.isfile(cache_path) and not self.force_reprocess:
            print(f"Loading cached UMLS data from {cache_path}")
            self.umls = pd.read_feather(cache_path)
            self.using_cached = True
            return

        print("Loading UMLS dataset.  This may take a few minutes")
        col_names = [
            "cui",
            "lang",
            "term_status",
            "lui",
            "stt",
            "sui",
            "ispref",
            "aui",
            "saui",
            "scui",
            "sdui",
            "sab",
            "tty",
            "code",
            "alias",
            "srl",
            "suppress",
            "cvf",
            "null_col",
        ]

        cols_to_use = [
            "cui",
            "sab",
            "sdui",
            "scui",
            "alias",
            "lang",
            "ispref",
            "tty",
            # "lui",
            # "sui",
            "aui",
            # "saui",
            "suppress",
        ]

        if self.debug:
            df = next(
                pd.read_csv(
                    os.path.join(self.umls_dir, "MRCONSO.RRF"),
                    delimiter="|",
                    names=col_names,
                    iterator=True,
                    # usecols=['cui','lang','sab','sdui','lui','sui','aui','saui','scui']
                    usecols=cols_to_use,
                    low_memory=False,
                    chunksize=10000,
                )
            )
        else:
            df = pd.read_csv(
                os.path.join(self.umls_dir, "MRCONSO.RRF"),
                delimiter="|",
                names=col_names,
                # usecols=['cui','lang','sab','sdui','lui','sui','aui','saui','scui']
                usecols=cols_to_use,
                low_memory=False,
                # engine='pyarrow',
            )

        print("Load rankings")
        rankings = pd.read_csv(
            os.path.join(self.umls_dir, "MRRANK.RRF"),
            delimiter="|",
            names=["rank", "sab", "tty", "_", "_1"],
            usecols=["rank", "sab", "tty"],
        )
        rankings["rank"] = -rankings["rank"] + 850

        print("Merge rankings")
        df = df.merge(rankings, on=["sab", "tty"])

        self.umls = df

    def _cache_umls(self):
        """
        Cache processed UMLS dataframe for faster reloading
        """
        if not self.debug and not self.using_cached:
            cache_path = os.path.join(self.umls_dir, ".cached_df.feather")
            print(f"Caching processed UMLS data to {cache_path}")
            print(self.umls.index)
            self.umls.reset_index(drop=True).to_feather(cache_path)

    def get_canonical_name(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        use_umls_curies: bool = True,
        mapping_cols: Optional[dict] = None,
        prefixes: Optional[dict] = None,
        remove_multigroup=False,
        reverse=False,
        lowercase=False,
    ):
        """
        Get canonical name of entities in UMLS Metathesaurus
        """
        df = self.filter_ontologies_and_types(
            ontologies_to_include=ontologies_to_include,
            types_to_include=types_to_include,
            groups_to_include=groups_to_include,
            remove_multigroup=remove_multigroup,
        )

        # Pick which set of CURIEs to use
        if use_umls_curies:
            df["identifier"] = df["cui"]
            filtered = df[["identifier", "rank", "alias"]]

        else:
            subsets = []
            for ontology in df.sab.unique():
                mapping_col = ""
                prefix = ""
                if ontology in mapping_cols:
                    mapping_col = mapping_cols[ontology]
                if ontology in prefixes:
                    prefix = prefixes[ontology]

                subset = self._get_alias_single_subset(
                    df, ontology, mapping_col=mapping_col, prefix=prefix
                )
                subsets.append(subset[["identifier", "rank", "alias"]])
            filtered = pd.concat(subsets).drop_duplicates()

        if lowercase:
            filtered["alias"] = filtered.alias.map(lambda x: x.lower())
            filtered = filtered.drop_duplicates()

        # Get output dict of CUI to Name mappings
        output_dict = (
            filtered.loc[filtered.groupby("identifier")["rank"].idxmin()]
            .set_index("identifier")["alias"]
            .to_dict()
        )
        return output_dict

    def get_definition(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        use_umls_curies: bool = True,
        mapping_cols: Optional[dict] = None,
        prefixes: Optional[dict] = None,
        remove_multigroup=False,
        reverse=False,
        lowercase=False,
    ):
        """
        Get canonical name of entities in UMLS Metathesaurus
        """
        print("Filtering by ontologies")
        df = self.filter_ontologies_and_types(
            ontologies_to_include=ontologies_to_include,
            types_to_include=types_to_include,
            groups_to_include=groups_to_include,
            remove_multigroup=remove_multigroup,
        )

        print("Removing null definitions")
        df = df.loc[~df['def'].isnull(),:]

        # Pick which set of CURIEs to use
        if use_umls_curies:
            df["identifier"] = df["cui"]
            filtered = df[["identifier", "rank", "def"]]

        else:
            subsets = []
            for ontology in df.sab.unique():
                mapping_col = ""
                prefix = ""
                if ontology in mapping_cols:
                    mapping_col = mapping_cols[ontology]
                if ontology in prefixes:
                    prefix = prefixes[ontology]

                subset = self._get_alias_single_subset(
                    df, ontology, mapping_col=mapping_col, prefix=prefix
                )
                subsets.append(subset[["identifier", "rank", "def"]])
            filtered = pd.concat(subsets).drop_duplicates()

        if lowercase:
            filtered["def"] = filtered['def'].map(lambda x: x.lower())
            filtered = filtered.drop_duplicates()

        # Get output dict of CUI to Name mappings
        output_dict = (
            filtered.loc[filtered.groupby("identifier")["rank"].idxmin()]
            .set_index("identifier")["def"]
            .to_dict()
        )
        return output_dict

    def get_types_and_groups(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        use_umls_curies: bool = True,
        mapping_cols: Optional[dict] = None,
        prefixes: Optional[dict] = None,
        remove_multigroup=False,
        reverse=False,
        lowercase=False,
    ):

        df = self.filter_ontologies_and_types(
            ontologies_to_include=ontologies_to_include,
            types_to_include=types_to_include,
            groups_to_include=groups_to_include,
            remove_multigroup=remove_multigroup,
        )

        # Pick which set of CURIEs to use
        if use_umls_curies:
            df["identifier"] = df["cui"]
            filtered = df[["identifier", "rank", "tui", "group"]]

        else:
            subsets = []
            for ontology in df.sab.unique():
                mapping_col = ""
                prefix = ""
                if ontology in mapping_cols:
                    mapping_col = mapping_cols[ontology]
                if ontology in prefixes:
                    prefix = prefixes[ontology]

                subset = self._get_alias_single_subset(
                    df, ontology, mapping_col=mapping_col, prefix=prefix
                )
                subsets.append(subset[["identifier", "rank", "tui", "group"]])
            filtered = pd.concat(subsets)

        grouped = filtered.groupby('identifier')
        id2types = grouped.tui.first().to_dict()
        id2groups = grouped.group.first().to_dict()

        return id2types, id2groups

    def get_mapping(
        self,
        ontology_abbreviation="MSH",
        reverse=False,
        mapping_col="sdui",
        umls_prefix="",
        other_prefix="",
    ):
        """
        Get cross-references between UMLS and another ongology
        """
        # Load main UMLS file (MRCONSO.RRF)

        # Filter down to desired ontology
        filtered = (
            self.umls.query("sab == @ontology_abbreviation")
            .loc[:, ["cui", mapping_col]]
            .dropna()
            .drop_duplicates()
            .reset_index(drop=True)
        )
        if umls_prefix != "":
            filtered["cui"] = filtered["cui"].map(lambda x: umls_prefix + ":" + x)

        if other_prefix != "":
            filtered[mapping_col] = filtered[mapping_col].map(
                lambda x: other_prefix + ":" + x
            )

        if reverse:
            output = filtered.set_index(mapping_col).to_dict()["cui"]
        else:
            output = filtered.set_index("cui").to_dict()[mapping_col]

        return output

    def interalias_mapping(self, type_subset="gngm"):
        """
        Get mapping between all terms with the same alias
        """
        pass
        # if hasattr(self.cui2type):

    def _load_types(self):
        """
        Load mapping from CUI to UMLS semantic type
        """
        # Load type dataframe
        print("Loading semantic types")
        cui_type_df = pd.read_csv(
            os.path.join(self.umls_dir, "MRSTY.RRF"),
            usecols=[0, 1],
            names=["cui", "tui"],
            delimiter="|",
            # engine='pyarrow'
        )

        self._load_type_abbrevs()

        cui_type_df["group"] = cui_type_df["tui"].map(self.tui2group)

        grouped = cui_type_df.groupby("cui")

        # Get mapping from CUI to types
        print("Mapping CUIs to types")
        self.cui2types = grouped["tui"].progress_apply(list).to_dict()

        # Get mapping from CUI to groups
        print("Mapping CUIs to groups")
        self.cui2groups = (
            grouped["group"].progress_apply(lambda x: list(set(x))).to_dict()
        )

    def _load_type_abbrevs(self):
        """
        Get type abbreviations and semantic groups
        """
        # Map each TUI to a semantic group
        self.tui2group = (
            pd.read_csv(
                os.path.join(self.umls_dir, "SemGroups.txt"),
                sep="|",
                usecols=[0, 2],
                names=["group", "tui"],
            )
            .set_index("tui")
            .to_dict()["group"]
        )

        self.semantic_groups = list(set(self.tui2group.values()))

        # Mape each TUI to its abbreviaion/name
        # Semantic network abbreviations
        sem_network_cols = [
            "row_type",
            "tui",
            "name",
            "tree_index",
            "desc",
            "_1",
            "clarification",
            "_2",
            "abbrev",
            "inverse_relation",
            "_3",
        ]

        self.tui_type_abbreviations = pd.read_csv(
            os.path.join(self.umls_dir, "semantic_network/SRDEF"),
            sep="|",
            names=sem_network_cols,
            usecols=["row_type", "tui", "name", "desc", "abbrev"],
        )
        self.type2abbrev = (
            self.tui_type_abbreviations[["tui", "abbrev"]]
            .set_index("tui")
            .to_dict()["abbrev"]
        )

    def _merge_types(self):
        """
        Merge type information into UMLS dataframe
        """
        print("Adding type information to UMLS dataframe")
        self.umls["tui"] = self.umls["cui"].map(self.cui2types)

        print("Adding semantic group informaiton to UMLS dataframe")
        self.umls["group"] = self.umls["cui"].map(self.cui2groups)

    def _load_type2name(self, umls_dir):
        """
        Load mapping from semantic types to names/abbreviations
        """
        pass

    def _get_alias_single_subset(self, df, ontology, mapping_col="", prefix=""):
        """
        Get CURIE -> alias mapping for a single subset of UMLS
        """
        filtered = df.query("sab == @ontology")
        inds = filtered.loc[:, [mapping_col, "alias"]].dropna().drop_duplicates().index
        filtered = filtered.loc[inds, :].reset_index(drop=True)
        filtered["identifier"] = filtered[mapping_col].map(lambda x: prefix + ":" + x)
        return filtered


    def list_groups(self):
        """
        List available semantic groups
        """
        return self.semantic_groups

    def filter_ontologies_and_types(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        remove_multigroup=False,
    ):

        # Pull correct subset of UMLS, if subsetting at all
        # print(self.umls.columns)
        if ontologies_to_include == "all":
            df = self.umls.query('lang == "ENG"')
        elif type(ontologies_to_include) == str:
            df = self.umls.query('sab == @ontologies_to_include & lang == "ENG"')
        else:
            df = self.umls[self.umls.sab.isin(ontologies_to_include)].query(
                'lang == "ENG"'
            )

        # Remove concepts with more than one group
        if remove_multigroup:
            df = df[df["group"].map(lambda x: len(x) == 1)]

        # Filter by semantic group
        if type(groups_to_include) == str:
            if groups_to_include != "all":
                raise ValueError('groups_to_include must be a list of strings or "all"')
        elif type(groups_to_include) == list:
            mask = df["group"].map(lambda x: any([y in groups_to_include for y in x]))
            df = df[mask]
        else:
            raise TypeError("groups_to_include must be list or string")

        # Filter by semantic type
        if type(types_to_include) == str:
            if types_to_include != "all":
                raise ValueError('types_to_include must be a list of strings or "all"')
        elif type(types_to_include) == list:
            mask = df["tui"].map(lambda x: any([y in types_to_include for y in x]))
            df = df[mask]
        else:
            raise TypeError("types_to_include must be list or string")

        return df

    def get_types(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        remove_multigroup=False,
    ):
        """
        Get semantic types for UMLS entities
        """

        df = self.filter_ontologies_and_types(
            ontologies_to_include=ontologies_to_include,
            types_to_include=types_to_include,
            groups_to_include=groups_to_include,
            remove_multigroup=remove_multigroup,
        )

        return df.groupby("cui").tui.first().to_dict()

    def get_aliases(
        self,
        ontologies_to_include: Union[str, list] = "all",
        types_to_include: Union[str, list] = "all",
        groups_to_include: Union[str, list] = "all",
        use_umls_curies: bool = True,
        mapping_cols: Optional[dict] = None,
        prefixes: Optional[dict] = None,
        remove_multigroup=False,
        reverse: bool = False,
        lowercase: bool = False,
    ):
        """
        Get mapping of CURIE -> aliases for all or a subset of UMLS entities

        Parameters:
        -----------------------------
            ontologies_to_include:
                Ontologies in UMLS to include in subset, default 'all'

            types_to_include:
                Semantic types to include in subset, default 'all'

            groups_to_include:
                Semantic groups to include in subset, default: 'all'
                Semantic groups are more general than types and typicially
                include multiple types.

            use_umls_curies: bool, default True
                Whether to use UMLS CUIs as unique identifiers.
                If false, use CUIs subset ontologies.

            mapping_cols:
                Dict of which column to use for CURIEs for each ontology.  If none
                specified, sdui is used.

            prefixes:
                Dict of prefixes to use for each ontology.  If none given, no prefixes
                are used

            reverse: bool, default False
                If true, return mapping from aliases to CURIEs instead of vice versa

            lowercase: bool, default False
                If true, lowercase all aliases before deduplicating/returning

            remove_multigroup: bool, default True
                If true, remove entities that belong to multiple groups
        """
        df = self.filter_ontologies_and_types(
            ontologies_to_include=ontologies_to_include,
            types_to_include=types_to_include,
            groups_to_include=groups_to_include,
            remove_multigroup=remove_multigroup,
        )

        # Pick which set of CURIEs to use
        if use_umls_curies:
            filtered = (
                df.loc[:, ["cui", "alias"]]
                .dropna()
                .drop_duplicates()
                .reset_index(drop=True)
            ).rename({"cui": "identifier"}, axis=1)

        else:
            subsets = []
            for ontology in df.sab.unique():
                mapping_col = ""
                prefix = ""
                if ontology in mapping_cols:
                    mapping_col = mapping_cols[ontology]
                if ontology in prefixes:
                    prefix = prefixes[ontology]

                subset = self._get_alias_single_subset(
                    df, ontology, mapping_col=mapping_col, prefix=prefix
                )
                subsets.append(subset[["identifier", "alias"]])
            filtered = pd.concat(subsets).drop_duplicates()

        if lowercase:
            filtered["alias"] = filtered.alias.map(lambda x: x.lower())
            filtered = filtered.drop_duplicates()

        if reverse:
            return filtered.groupby("alias")["identifier"].apply(list).to_dict()
        else:
            return filtered.groupby("identifier")["alias"].apply(list).to_dict()

    def get_synsets(self, preferred_type="gngm"):
        """
        Get mapping between all terms with the same alias

        This is mostly to get mapping from proteins to their corresponding gene and vice-versa
        """
        df = self.umls
        df["type"] = df.tui.map(self.type2abbrev)

        df["has_preferred_type"] = df["type"] == preferred_type

        grouped = self.umls.groupby("alias")


# Semantic network abbreviations
# def get_semantic_type_hierarchy(umls_dir, tui):
#     sem_network_cols = [
#         "row_type",
#         "tui",
#         "name",
#         "tree_index",
#         "desc",
#         "_1",
#         "clarification",
#         "_2",
#         "abbrev",
#         "inverse_relation",
#         "_3",
#     ]

#     semantic_network = pd.read_csv(
#         os.path.join(umls_dir, "semantic_network/SRDEF"),
#         sep="|",
#         names=sem_network_cols,
#         usecols=["row_type", "tui", "name", "tree_index"],
#     )

#     tree_index_to_name = semantic_network.set_index('tree_index')['name'].to_dict()


# def read_sem_group():
#     pass


# def download_umls():
#     pass



import warnings
import os
import ujson
from collections import defaultdict
import pandas as pd
from datasets import load_dataset

from bioel.utils.dataset_consts import *
from bioel.logger import setup_logger

logger = setup_logger()


def load_bigbio_dataset(dataset_name):
    """
    Load BigBio dataset and include abbreviations if specified.

    Params
    ------
    - dataset_name : str
        Name of the dataset to load
    """
    # Load the dataset
    if dataset_name in {"medmentions_st21pv", "medmentions_full"}:
        dataset = load_dataset(
            f"bigbio/medmentions",
            name=f"{dataset_name}_bigbio_kb",
            trust_remote_code=True,
        )
    else:
        dataset = load_dataset(
            f"bigbio/{dataset_name}",
            name=f"{dataset_name}_bigbio_kb",
            trust_remote_code=True,
        )

    return dataset


def add_deabbreviations(dataset, path_to_abbrev):
    """
    Function to add deabbreviated mentions to the entities in the dataset

    Params
    ------
    - dataset : DatasetDict
        The dataset to which deabbreviations need to be added
    - path_to_abbrev : str
        Path to the JSON file containing abbreviations
    """

    if not os.path.exists(path_to_abbrev):
        raise FileNotFoundError(
            f"The file {path_to_abbrev} does not exist. \nCreate the abbreviations.json file first using create_abbrev function from solve_abbreviations.py"
        )

    with open(path_to_abbrev, "r") as f:
        abbreviations = ujson.load(f)

    for split in dataset.keys():
        dataset[split] = dataset[split].map(
            lambda example: {
                "entities": [
                    {
                        **entity,
                        "deabbreviated_text": resolve_abbreviation(
                            example["document_id"], entity["text"], abbreviations
                        ),
                    }
                    for entity in example["entities"]
                ]
            },
            batched=False,
        )

    return dataset


def cache_deduplicated_dataset(deduplicated_df):
    """
    Cache deduplicated dataset file for faster loading
    """
    raise NotImplementedError


def check_if_cached(dataset_name):
    """
    Check if dataset has been preprocessed and cached
    """
    raise NotImplementedError


def load_cached_dataset(dataset, splits_to_include: list = None):
    """
    Load cached deduplciated dataset as dataframe
    """
    raise NotImplementedError


def dataset_to_documents(dataset):
    """
    Return dictionary of documents in BigBio dataset
    """
    docs = {}

    for split in dataset.keys():
        for doc in dataset[split]:
            doc_id = pmid = doc["document_id"]
            doc_text = "\n".join([" ".join(x["text"]) for x in doc["passages"]])
            docs[doc_id] = doc_text
    return docs


def get_dataset_documents(dataset_name):
    """
    Get the documents from a BigBio dataset
    """
    dataset = load_bigbio_dataset(dataset_name)
    return dataset_to_documents(dataset)


def make_mention_id(document_id, running_mention_count):
    """
    Make a unique ID for each mention
    """
    running_mention_count[document_id] += 1
    return f"{document_id}.{running_mention_count[document_id]}"


# def get_dataset_df(dataset_name, abbrev_dict: dict = None):
#     data = load_bigbio_dataset(dataset_name)

#     data_df = dataset_to_df(
#         data,
#         entity_remapping_dict=CUIS_TO_REMAP[dataset_name],
#         cuis_to_exclude=CUIS_TO_EXCLUDE[dataset_name],
#         val_split_ids=VALIDATION_DOCUMENT_IDS[dataset_name],
#     )

#     if abbrev_dict is not None:
#         data_df["deabbreviated_text"] = data_df[["document_id", "text"]].apply(
#             lambda x: resolve_abbreviation(
#                 document_id=x[0],
#                 text=x[1],
#                 abbreviations_dict=abbrev_dict,
#             ),
#             axis=1,
#         )

#     return data_df


# Function to match text with abbreviations for each entity
def match_abbreviation(entity, doc_id, abbrev_dict):
    doc_abbrevs = abbrev_dict.get(str(doc_id), {})
    for text in entity["text"]:
        for abbr, full_form in doc_abbrevs.items():
            if full_form == text:
                return abbr
    return None


def dataset_to_df(
    dataset,
    splits_to_include: list = None,
    entity_remapping_dict: dict = None,
    cuis_to_exclude: list = None,
    val_split_ids: list = None,
    # abbreviations_dict: dict = None,
):
    """
    Convert BigBio dataset to pandas DataFrame

    Params:
    ------------------
        dataset: BigBio Dataset
            Dataset to load from BigBio

        splits_to_include: list of str
            List of splits to include in mo
    """
    columns = [
        # 'context', # string
        "document_id",  # string
        "mention_id",  # string
        "text",  # string
        "type",  # list
        "offsets",  # list of lists
        # "db_name",
        "db_ids",  # list
        "split",  # string
        # "abbreviation_resolved", # bool
        "deabbreviated_text",  # string
    ]
    all_lines = []

    if splits_to_include is None:
        splits_to_include = dataset.keys()

    for split in splits_to_include:
        if split not in dataset.keys():
            warnings.warn(f"Split '{split}' not in dataset.  Omitting.")
        for doc in dataset[split]:
            pmid = doc["document_id"]
            for e in doc["entities"]:
                if len(e["normalized"]) == 0:
                    continue
                text = " ".join(e["text"])
                # abbreviation_resolved = False
                offsets = ";".join(
                    [",".join([str(y) for y in x]) for x in e["offsets"]]
                )
                # db_name = e["normalized"][0]["db_name"]
                db_ids = [
                    x["db_name"] + ":" + x["db_id"].strip() for x in e["normalized"]
                ]

                # Get the abbreviation if it exists, else set to None or an empty string
                deabbreviated_text = e.get("deabbreviated_text", None)

                # Remap entity IDs when identifier has changed in database
                if entity_remapping_dict is not None:
                    db_ids = [
                        entity_remapping_dict[x] if x in entity_remapping_dict else x
                        for x in db_ids
                    ]

                # Remove any IDs not included in database
                # Remove mentions with no entity link in database
                if cuis_to_exclude is not None:
                    new_db_ids = [x for x in db_ids if x not in cuis_to_exclude]
                else:
                    new_db_ids = db_ids
                if len(new_db_ids) == 0:
                    continue

                # Add mention + metadata to list of mentions
                all_lines.append(
                    [
                        pmid,
                        e["id"],
                        text,
                        e["type"],
                        # e['offsets'],
                        offsets,
                        # db_name,
                        new_db_ids,
                        split,
                        # abbreviation_resolved,
                        deabbreviated_text,
                    ]
                )

    df = pd.DataFrame(all_lines, columns=columns)

    deduplicated = (
        df.groupby(["document_id", "offsets"])
        .agg(
            {
                "text": "first",
                "type": lambda x: list(set([a for a in x])),
                "db_ids": lambda db_ids: list(set([y for x in db_ids for y in x])),
                "split": "first",
                "deabbreviated_text": "first",
            }
        )
        .reset_index()
    )

    deduplicated["offsets"] = deduplicated["offsets"].map(
        lambda x: [[int(z) for z in y.split(",")] for y in x.split(";")]
    )

    # Order mentions in consistent way (i.e. )
    deduplicated["first_offset"] = deduplicated["offsets"].map(lambda x: x[0][0])
    deduplicated["last_offset"] = deduplicated["offsets"].map(lambda x: x[-1][-1])
    deduplicated = deduplicated.sort_values(
        by=[
            "document_id",
            "first_offset",
            "last_offset",
        ]
    )
    deduplicated = deduplicated.drop(["first_offset", "last_offset"], axis=1)

    # Make a unique mention ID for each mention
    mention_counts = defaultdict(int)
    deduplicated["mention_id"] = deduplicated["document_id"].map(
        lambda x: make_mention_id(x, running_mention_count=mention_counts)
    )

    # Split off validation set if not given
    if val_split_ids is not None:
        print(type(val_split_ids[0]), type(deduplicated["document_id"][0]))
        deduplicated.loc[deduplicated["document_id"].isin(val_split_ids), "split"] = (
            "validation"
        )

    return deduplicated


def get_left_context(doc, start, max_length=64, strip=False):
    """
    Get all text in document that comes after a mentions up to $max_len words
    """
    if strip:
        return " ".join(doc[:start].strip().split()[-max_length:])
    else:
        return " ".join(doc[:start].split()[-max_length:])


def get_right_context(doc, end, max_length=64, strip=False):
    """
    Get all text in document that comes after a mentions up to $max_len words
    """
    if strip:
        return " ".join(doc[end:].strip().split()[:max_length])
    else:
        return " ".join(doc[end:].split()[:max_length])


def resolve_abbreviation(doc_id, texts, abbreviations):
    """
    Resolve abbreviations for a given document and list of texts using the abbreviations dictionary.

    Params
    ------
    - doc_id : str
        Document ID to look up in abbreviations
    - texts : list of str
        List of texts to resolve abbreviations for
    - abbreviations : dict
        Dictionary containing abbreviation mappings

    Returns
    -------
    - list of str
        List of resolved abbreviations if found, else original texts
    """
    resolved_texts = [abbreviations.get(doc_id, {}).get(text, text) for text in texts]
    return "".join(resolved_texts)


def load_dataset_df(name, path_to_abbrev=None):
    """
    Load bigbio dataset and turn into pandas dataframe

    Params
    ------
    - name : str
        Name of the dataset to load
    - path_to_abbrev : str
        Path to the JSON file containing abbreviations
    """
    data = load_bigbio_dataset(dataset_name=name)

    if path_to_abbrev:
        data = add_deabbreviations(dataset=data, path_to_abbrev=path_to_abbrev)

    if name in CUIS_TO_EXCLUDE:
        exclude = CUIS_TO_EXCLUDE[name]
    else:
        exclude = None

    if name in CUIS_TO_REMAP:
        remap = CUIS_TO_REMAP[name]
    else:
        remap = None

    if name in VALIDATION_DOCUMENT_IDS:
        validation_pmids = VALIDATION_DOCUMENT_IDS[name]
    else:
        validation_pmids = None

    df = dataset_to_df(
        data,
        entity_remapping_dict=remap,
        cuis_to_exclude=exclude,
        val_split_ids=validation_pmids,
    )

    return df


def metamap_text_to_candidates(metamap_output):
    """
    Create mapping from text to list of candidates output by metamap

    Inputs:
    -------------------------------
        filepath: string or path-like
            Path to file containing output of MetaMap

    Returns:
    -------------------------------
        text2candidates: dict
            Dict of form {mention_text: [cand1, cand2, ...]} mapping each text string to candidates
            generated by metamap.  If no candidates were generated for a given key, value will be
            an empty string.
    """
    text2candidates = defaultdict(list)

    for row in metamap_output[
        ["text", "mapping_cui_list", "candidate_cui_list"]
    ].values:
        text = row[0]
        candidates = eval(row[1]) + eval(row[2])

        # TODO: Need to account for correct database
        raise NotImplementedError("Need to map UMLS values to correct DB")
        candidates = ["UMLS:" + x for x in candidates]

        text2candidates[text] = candidates
    return text2candidates


def dataset_unique_tax_ids(dataset: str, entrez):
    """
    Return a list of unique tax_ids from a dataset

    Parameters
    ----------
    dataset: str
        Name of the dataset to load
    """
    data = load_bigbio_dataset(dataset_name=dataset)
    exclude = CUIS_TO_EXCLUDE[dataset]
    remap = CUIS_TO_REMAP[dataset]

    df = dataset_to_df(data, entity_remapping_dict=remap, cuis_to_exclude=exclude)

    # Flatten the db_ids lists into a single list
    all_db_ids = [db_id for sublist in df["db_ids"] for db_id in sublist]

    # Count the number of unique db_ids
    unique_db_ids = set(all_db_ids)

    unique_gene_ids = [
        int(gene_id.replace("NCBIGene:", "")) for gene_id in unique_db_ids
    ]

    unique_tax_id = (
        entrez[entrez["geneid"].isin(unique_gene_ids)]["tax_id"]
        .drop_duplicates()
        .tolist()
    )

    logger.info(f"Number of unique tax_ids in {dataset} : {len(unique_tax_id)}")

    return unique_tax_id


if __name__ == "__main__":
    dataset_name = "bc5cdr"
    dataset = load_bigbio_dataset(dataset_name)
    print(dataset_to_documents(dataset))